from pathlib import Path
import logging
logger = logging.getLogger(Path(__file__).stem)

def 解鎖(pdfs, 覆蓋原檔=False):
    import PyPDF2
    from collections.abc import Iterable 
    from pathlib import Path
    if not isinstance(pdfs, Iterable):
        pdfs = [pdfs]
    for pdf in pdfs:
        with open(pdf, 'rb') as input_file:
            reader = PyPDF2.PdfReader(input_file)
            writer = PyPDF2.PdfWriter()
            for page_num in range(len(reader.pages)):
                writer.add_page(reader.pages[page_num])
            desecured_pdf:Path = pdf.with_stem(pdf.stem + '_desecured')
            with open(desecured_pdf, 'wb') as output_file:
                writer.write(output_file)
            if 覆蓋原檔:
                desecured_pdf.replace(pdf)


def 合併(pdfs, 合併檔名='合併檔案.pdf'):
    from zhongwen.office_document import doc2docx, doc2pdf
    from PyPDF2 import PdfWriter
    from pathlib import Path
    from shutil import copy
    pdfs = sorted(pdfs)
    merger = PdfWriter()
    filedir = None
    docxdir = None
    pdfdir = None
    for src in pdfs:
        src = Path(src)
        if not filedir:
            filedir = Path(src).parent
        if not docxdir:
            docxdir = Path(src).parent / 'docxs'
            docxdir.mkdir(exist_ok=True)
        if not pdfdir:
            pdfdir = Path(src).parent / 'pdfs'
            pdfdir.mkdir(exist_ok=True)
        if src.suffix == '.doc':
            doc2docx(src, docxdir)
            docx = docxdir / src.with_suffix('.docx').name
            doc2pdf(docx, pdfdir)
            pdf = pdfdir / docx.with_suffix('.pdf').name
        if src.suffix == '.docx':
            doc2pdf(src, pdfdir)
            pdf = pdfdir / src.with_suffix('.pdf').name
        if src.suffix == '.pdf':
            copy(src, pdfdir)
        try:
            merger.append(pdf)
        except Exception as e:
           breakpoint() 
           pass
    output = filedir / 合併檔名
    merger.write(output)
    merger.close()

def 旋轉(源檔, 角度=180, 旋轉頁=None, 目標檔=None):
    import PyPDF2
    from pathlib import Path
    if not 目標檔: 
        目標檔 = 源檔.with_stem(源檔.stem+"旋轉後")
    with open(源檔, 'rb') as pdf_in, open(目標檔, 'wb') as pdf_out:
        pdf_reader = PyPDF2.PdfReader(pdf_in)
        pdf_writer = PyPDF2.PdfWriter()
        for pagenum in range(len(pdf_reader.pages)):
            page = pdf_reader.pages[pagenum]
            if pagenum in 旋轉頁:
                page.rotate(角度)
            pdf_writer.add_page(page)
        pdf_writer.write(pdf_out)

def 去空白頁(源檔, 轉換檔=None):
    if not 轉換檔: 
        轉換檔 = 源檔.with_stem(源檔.stem+"去空白頁")
    input_file = 源檔 
    output_file = 轉換檔

    import fitz
    doc = fitz.open(input_file)

    非空白頁碼清單 = []
    import numpy as np 
    from PIL import Image
    for page_number in range(len(doc)):
        page = doc.load_page(page_number)
        pix = page.get_pixmap()
        
        # 将Pixmap对象转换为PIL的Image对象
        image = Image.frombytes("RGB", [pix.width, pix.height], pix.samples)
        
        # 将图像转换为灰度图像
        gray_image = image.convert("L")

        # 将灰度图像转换为NumPy数组
        np_image = np.array(gray_image)
        
        # 计算像素值的均值
        mean_value = np.mean(np_image)
        
        # 设置阈值来判断页面是否为空白
        threshold = 200  # 可根据需要进行调整

        if mean_value < threshold: 
            非空白頁碼清單.append(page_number)
    doc.select(非空白頁碼清單)
    doc.save(output_file)

def 刪除頁面(input_path, pages, output_path=None):
    from PyPDF2 import PdfWriter, PdfReader
    from pathlib import Path
    input_path = Path(input_path)
    if not output_path:
        s = input_path.stem
        output_path = input_path.with_stem(f'{s}刪除後')
    with open(input_path, 'rb') as file:
        reader = PdfReader(file)
        writer = PdfWriter()

        total_pages = len(reader.pages)

        for page_number in range(total_pages):
            if page_number + 1 not in pages:
                writer.add_page(reader.pages[page_number])

        with open(output_path, 'wb') as output_file:
            writer.write(output_file)
        print(f"{input_path}刪除頁面{', '.join(map(str, pages))}，結果另存於{output_path}。")

def to_excel(pdfs):
    from collections.abc import Iterable 
    import pandas as pd
    import pdfplumber
    if isinstance(pdfs, str) or not isinstance(pdfs, Iterable):
        pdfs = [pdfs]
    
    for pdf in pdfs:
        with pdfplumber.open(pdf) as _pdf:
            all_data = []
            for page in _pdf.pages:
                table = page.extract_table()
                if table:
                    df = pd.DataFrame(table[1:], columns=table[0])  # 使用表格第一行作為標題
                    all_data.append(df)
            final_df = pd.concat(all_data)
            xlsx = pdf.with_suffix('.xlsx')
            final_df.to_excel(xlsx, index=False)
            print(f'{pdf.name}->{xlsx.name}')


def 設定環境():
    from zhongwen.winman import 建立傳送到項目
    import sys
    cmd =  f'"{sys.executable}" -m zhongwen.pdf --merge_pdfs %* && pause'
    建立傳送到項目('合併為PDF', cmd)

if __name__ == '__main__':
    import argparse
    logging.basicConfig(level=logging.INFO)
    parser = argparse.ArgumentParser()
    parser.add_argument("--setup", help="設定環境", action='store_true')
    parser.add_argument('--merge_pdfs', nargs='+', type=str, help='合併PDF')
    args = parser.parse_args()
    if args.setup:
        設定環境()
    elif pdfs := args.merge_pdfs:
        合併(pdfs)




import unittest

class Test(unittest.TestCase):
    def test_take_periods(self):
        from zhongwen.時 import 取期間
        from pandas import Period, Timestamp
        self.assertEqual(取期間('11310'), Period('202410', 'M'))
        self.assertEqual(取期間('113/10'), Period('202410', 'M'))
        self.assertEqual(取期間('113/1'), Period('202401', 'M'))
        self.assertEqual(取期間('2024-10'), Period('202410', 'M'))
        self.assertEqual(取期間('2020-01'), Period('202001', 'M'))

    def test_period(self):
        from zhongwen.時 import 上月
        import pandas as pd
        self.assertIsInstance(上月(), pd.Period)

    def test_iter_month(self):
        from zhongwen.時 import 自指定月份迄上月
        import pandas as pd
        self.assertEqual(list(自指定月份迄上月('11301')), 
                         list(pd.period_range('2024-01', '2024-10', freq='M'))
                        )

if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.INFO)
    logging.getLogger('googleclient').setLevel(logging.CRITICAL)
    logging.getLogger('matplotlib').setLevel(logging.CRITICAL)
    logging.getLogger('faker').setLevel(logging.CRITICAL)
    unittest.main()
    suite = unittest.TestSuite()
    suite.addTest(Test('test_take_period'))  # 指定測試
    unittest.TextTestRunner().run(suite)



from zhongwen.date import 取日期
from zhongwen.date import 今日, 年底, 上年底
from zhongwen.date import 民國日期, 民國年月
from zhongwen.date import 自起日按日列舉迄今

def 正式民國日期(d=None):
    '格式如：112年7月29日'
    if not d:
        d = 今日()
    return 民國日期(d, "%Y年%M月%D日")

def 上月():
    import pandas as pd
    return pd.Period(今日(), 'M') - 1

def 自指定月份迄上月(月份):
    from pandas import Period
    月份 = 取期間(月份)
    while 月份 <= 上月():
        yield 月份
        月份+=1
    return 上月()

def 取期間(期間字串, 全取=False):
    '指定全取則回傳期間串列，否則傳為首個期間'
    from pandas import Period
    import re

    s = 期間字串
    if ms:=re.findall(r'(\d{4})-([01]?\d)', s):
        ps = [Period(f'{int(m[0])}{int(m[1]):02}', 'M') for m in ms]
    elif ms:=re.findall(r'(\d{3})[/]?([01]?\d)', s):
        try:
            ps = [Period(f'{int(m[0])+1911}{int(m[1]):02}', 'M') for m in ms]
        except Exception:
            raise Exception(f"'{期間字串}'無法解析為期間字串！")
   
    if 全取: return ps
    
    try:
        return ps[0]
    except UnboundLocalError:
        raise UnboundLocalError(f"'{期間字串}'無法解析為期間字串！")



import unittest

class Test(unittest.TestCase):
    def test(self):
        from zhongwen.自動完成 import 同字首詞
        行 = '是以，依據美國十年'
        字典 = ['公債',  '十年房地產', '美國十年期公債', '美國十年經濟'] 
        結果 = [{'abbr': '美國十年期公債', 'kind': 'K', 'word': '期公債'}
               ,{'abbr': '美國十年經濟', 'kind': 'K', 'word': '經濟'}
               ,{'abbr': '十年房地產', 'kind': 'K', 'word': '房地產'}]
        ['美國十年期公債', '美國十年經濟', '十年房地產'] 
        self.assertEqual(同字首詞(行, 字典), 結果)


if __name__ == '__main__':
    unittest.main()



'輔助PANDAS工具'
from pathlib import Path
import logging
logger = logging.getLogger(Path(__file__).stem)

class 使用者要求更新且線上資料較線下多(Exception): pass
class 使用者要求覆寫(Exception): pass

def 增加自動更新資料功能(更新頻率='每月十日之前'):
    '如將參數「更新」設為True，則強制更新該批資料'
    import logging
    from pathlib import Path
    from diskcache import Cache
    from functools import wraps
    from zhongwen.date import 今日 
    cache = Cache(Path.home() / 'cache' / Path(__file__).stem)
    def _增加按期更新查詢結果功能(查詢資料):
        @wraps(查詢資料)
        def 查詢按期更新資料(*args, 更新=False,**kargs):
            if 更新頻率=='每月十日之前':
                if 今日().day <= 10 or 更新:
                    df = 查詢資料(*args, **kargs)
                    cache.set(查詢資料.__name__, df)
                    return df
            return cache.read(查詢資料.__name__)
        return 查詢按期更新資料
    return _增加按期更新查詢結果功能

class 應更新(Exception): pass

def 增加按期更新查詢結果功能(更新頻率='每月十日之前'):
    '如將參數「更新」設為True，則強制更新該批資料'
    import logging
    from pathlib import Path
    from diskcache import Cache
    from zhongwen.date import 今日 
    warn(f'【增加按期更新查詢結果功能】將廢棄，請使用【batch_data.增加定期更新】', DeprecationWarning, stacklevel=2)
    cache = Cache(Path.home() / 'cache' / Path(__file__).stem)
    def _增加按期更新查詢結果功能(查詢資料):
        from functools import wraps
        @wraps(查詢資料)
        def 查詢按期更新資料(*args, 更新=False,**kargs):
            try:
                if 更新 or (更新頻率=='每月十日之前' and 今日().day <= 10): raise 應更新()
                return cache.read(查詢資料.__name__)
            except (KeyError, 應更新) as e:
                df = 查詢資料(*args, **kargs)
                cache.set(查詢資料.__name__, df)
                return df
        return 查詢按期更新資料
    return _增加按期更新查詢結果功能

def 增加批次儲存(資料庫檔, 資料名稱, 批號欄名):
    from functools import wraps
    from sqlite3 import connect
    import pandas as pd
    import logging
    from warnings import warn
    warn(f'【增加批次緩存】將廢棄，請使用【batch_data.結果批次寫入】', DeprecationWarning, stacklevel=2)
    def 增加批次儲存功能(資料存取函數):
        @wraps(資料存取函數)
        def 批次資料存取(批號, *args, **kargs):
            df = 資料存取函數(批號, *args, **kargs)
            with connect(資料庫檔) as db: 
                批次寫入(df, 批號, 批號欄名, 資料名稱, db)
            return df
        return 批次資料存取
    return 增加批次儲存功能

def 增加批次緩存功能(資料庫檔, 資料名稱, 批號欄名):
    '如將參數「更新」設為True，則強制更新該批資料'
    from functools import wraps
    from sqlite3 import connect
    import pandas as pd
    import logging
    from warnings import warn
    warn(f'【增加批次緩存功能】將廢棄，請使用【batch_data.結果批次寫入】', DeprecationWarning, stacklevel=2)
    def _增加批次緩存功能(資料存取函數):
        @wraps(資料存取函數)
        def 批次緩存資料存取(批號, *args, 更新=False, 覆寫=False,**kargs):
            with connect(資料庫檔) as db: 
                try:
                    df0 = 批次讀取(批號, 批號欄名, 資料名稱, db) 
                    if 更新:
                        df1 = 資料存取函數(批號, *args, **kargs)
                        if 覆寫: raise 使用者要求覆寫()
                        if df1.shape[0] > df0.shape[0]:
                            logging.info(
                                f'目前抓取{批號}之{資料名稱}計{df1.shape[0]}筆，'
                                f'已較緩存資料庫{df0.shape[0]}筆多，爰更新資料庫！')    
                            raise 使用者要求更新且線上資料較線下多()
                    return df0 
                except (pd.errors.DatabaseError, 批號查無資料錯誤) as e:
                    logging.info(e)
                    df1 = 資料存取函數(批號, *args, **kargs)
                    批次寫入(df1, 批號, 批號欄名, 資料名稱, db, 更新)
                    return df1
                except (使用者要求覆寫, 使用者要求更新且線上資料較線下多) as e:
                    logging.info(e)
                    批次寫入(df1, 批號, 批號欄名, 資料名稱, db, 覆寫=True)
                    return df1
        return 批次緩存資料存取
    return _增加批次緩存功能

class 批號存在錯誤(Exception):
    def __init__(self, 批號, 批號欄名, 表格):
        self.批號 = 批號
        self.批號欄名 = 批號欄名
        self.表格 = 表格

    def __str__(self):
        return f'批號【{self.批號}】已存在，請指定覆寫=True！'

class 批號查無資料錯誤(Exception): pass

def 批次讀取(批號, 批號欄名, 表格, 資料庫, parse_dates=None):
    import pandas as pd
    from zhongwen.date import 是日期嗎
    # sql 的相等係單等號表示，而python為雙等號，易誤植
    sql = f'select * from {表格} where {批號欄名}={批號}' 
    if 是日期嗎(批號) or isinstance(批號, str):
        sql = f'select * from {表格} where {批號欄名}="{批號}"' 
    df = pd.read_sql_query(sql, 資料庫, parse_dates=parse_dates)
    if df.empty: 
        raise 批號查無資料錯誤(f'查無批號為【{批號}】資料！')
    return df

def 批次刪除(批號, 批號欄名, 表格, 資料庫):
    c = 資料庫.cursor()
    sql = f'delete from {表格} where {批號欄名}=="{批號}"'
    c.execute(sql)
    資料庫.commit()
    import logging
    logging.debug(f'批次刪除{批號}、{表格}……')

def 批次寫入(資料, 批號, 批號欄名, 表格, 資料庫, 覆寫=False):
    from warnings import warn
    import logging
    import pandas as pd
    import sqlite3
    import re
    warn(f'參數【覆寫】將廢棄。', DeprecationWarning, stacklevel=2)
    try:
        批次刪除(批號, 批號欄名, 表格, 資料庫)
    except sqlite3.OperationalError as e:
        logging.debug(f'批次刪除發生錯誤{e}')
    logging.debug(f'批次寫入{批號}、{表格}……')
    try:
        資料.to_sql(表格, 資料庫, if_exists='append')
    except sqlite3.OperationalError as e:
        df = pd.read_sql_query(f'select * from {表格} limit 1', 資料庫)
        寫入資料欄位 = 資料.columns.to_list()
        資料庫欄位 = df.columns.to_list()
        差異欄位 = set(寫入資料欄位) - set(資料庫欄位)
        print(差異欄位)
        cursor = 資料庫.cursor()
        for c in 差異欄位:
            alter_query = f'ALTER TABLE {表格} ADD COLUMN "{c}" INT'
            cursor.execute(alter_query)
        資料庫.commit()
        print('已新增差異欄位')
        資料.to_sql(表格, 資料庫, if_exists='append')
        # raise e
    logging.debug(f'寫入成功！')

def 載入時間序列(資料庫檔, 表格, 時間欄名, 最早時間=None, 實體欄名=None):
    '由實體及時間組成索引鍵'
    import pandas as pd
    import sqlite3
    with sqlite3.connect(資料庫檔) as c:
        sql = f"select * from {表格}"
        if 最早時間:
            sql += f" where {時間欄名} >= '{最早時間:%Y-%m-%d}' order by {時間欄名}"
        df = pd.read_sql_query(sql, c, index_col='index', parse_dates=時間欄名) 
        df[實體欄名] = df[實體欄名].astype(str)
        return df.set_index([實體欄名, 時間欄名])

def 可顯示(查詢資料函數):
    '裝飾查詢資料函數，指名參數設為【顯示=True】，即將查詢結果以 html 顯示。'
    from functools import wraps
    import matplotlib.pyplot as plt
    @wraps(查詢資料函數)
    def wrapper(*args, 顯示=False, 圖示=False, **kargs):
        try:
            display_rows = kargs['顯示筆數']
            real_columns = kargs['實數欄位']
            del kargs['顯示筆數']
            del kargs['實數欄位']
        except KeyError:
            display_rows = 100
            real_columns = []
        df = 查詢資料函數(*args, **kargs)
        if 顯示: show_html(df, 顯示筆數=display_rows, 實數欄位=real_columns)
        if 圖示: 
            plt.rcParams['font.sans-serif'] = ['Microsoft JhengHei']
            df.plot()
            plt.show()
        return df
    return wrapper

def 強調關鍵字(df, 欄位=[], 關鍵字=[]):
    for c in 欄位:
        for k in 關鍵字:
            df[c] = df[c].str.replace(k, f'<b style="background:pink">{k}</b>')
    return df

def 資料型態標準化(df):
    '依資料框欄位名稱自動將資料框之資料型態標準化'
    from zhongwen.date import 取日期
    from zhongwen.number import 轉數值
    for c in df.columns:
        p = f'.*日期'
        import re
        if re.match(p, c):
            df[c] = df[c].map(取日期)
        p = f'.*金額'
        if re.match(p, c):
            df[c] = df[c].map(轉數值)
    return df            

def 標準格式(整數欄位=[], 實數欄位=[], 百分比欄位=[], 日期欄位=[], 
             最大值顯著欄位=[], 隱藏欄位=[], 
             百分比漸層欄位=[], 
             採用民國日期格式=False
            ):
    import logging
    logging.debug(f'整數欄位：{整數欄位!r}')
    logging.debug(f'實數欄位：{實數欄位!r}')
    logging.debug(f'百分比欄位：{百分比欄位!r}')
    logging.debug(f'百分比漸層欄位：{百分比漸層欄位!r}')
    logging.debug(f'最大值顯著欄位：{最大值顯著欄位!r}')
    logging.debug(f'日期欄位：{日期欄位!r}')
    logging.debug(f'隱藏欄位：{隱藏欄位!r}')

    def formatter(style):
        style.map(lambda r:'text-align:right')
        if 整數欄位:
            style.format('{:,.0f}', subset=整數欄位)
        if 百分比欄位:
            style.format('{:,.2%}', subset=百分比欄位)
        if 實數欄位:
            style.format('{:,.2f}', subset=實數欄位)
        if 日期欄位:
            style.format('{:%Y%m%d}', subset=日期欄位)
            if 採用民國日期格式:
                from zhongwen.date import 民國日期
                style.format(民國日期, subset=日期欄位)
        if 最大值顯著欄位:
            # style.highlight_max(最大值顯著欄位)
            style.background_gradient(axis=0, cmap='RdYlGn'
                                     ,subset=最大值顯著欄位)
        if 百分比漸層欄位:
            style.background_gradient(axis=0, cmap='RdYlGn', vmax=1, vmin=-1
                                     ,subset=百分比漸層欄位
                                     )
        if 隱藏欄位:
            style.hide(隱藏欄位, axis=1) # hide index
        tr_hover = {
            'selector': 'tr:hover',
            'props':[('background-color', '#ffffb3'), ('border-style', 'dotted')]
        }
        style.set_table_styles([tr_hover], overwrite=False)


        return style
    return formatter

def show_html(df, 無格式=False
             ,整數欄位=[], 實數欄位=[]
             ,百分比欄位=[] ,日期欄位=[] ,隱藏欄位=[]
             ,百分比漸層欄位=[]
             ,最大值顯著欄位=[]
             ,顯示筆數=100, 採用民國日期格式=False, 標題=None
             ,傳回超文件內容=False
             ):
    from pathlib import Path
    from warnings import warn
    import pandas as pd
    import numpy as np
    import os

    if isinstance(df, set):
        df = pd.Series(list(df)).to_frame()

    if isinstance(df, np.ndarray):
        df = pd.Series(df).to_frame()

    if isinstance(df, pd.Series):
        df = df.to_frame()

    # df = df.fillna(' ')
    if isinstance(df, pd.DataFrame):
        df = df.dropna(axis='columns', how='all')
        if not df.index.is_unique:
            df = df.reset_index(drop=True)

    if isinstance(df, pd.DataFrame) and not 無格式:
        try:
            df = df.head(顯示筆數)
            return 自動格式(df
                           ,整數欄位
                           ,實數欄位
                           ,百分比欄位
                           ,日期欄位
                           ,隱藏欄位
                           ,百分比漸層欄位
                           ,最大值顯著欄位
                           ,顯示筆數
                           ,採用民國日期格式
                           ,顯示=True
                           )
        except (ValueError, KeyError, TypeError) as e: 
            import traceback
            trace = traceback.format_exc()
            warn(f'發生例外：{trace}')
    html = Path.home() / 'TEMP' / 'output.html'
    df.to_html(html)
    os.system(f'start {html}')


def 自動格式(df, 整數欄位=[] ,實數欄位=[], 百分比欄位=[]
            ,日期欄位=[] ,隱藏欄位=[]
            ,百分比漸層欄位=[], 最大值顯著欄位=[], 不排序欄位=[]
            ,顯示筆數=100, 採用民國日期格式=False, 顯示=None
            ,除錯提示=False, 顯示提示=True
            ):
    from zhongwen.number import 轉數值
    import pandas as pd
    import numpy as np
    import re
    # df = df.fillna('')
    if 顯示筆數:
        df = df[:顯示筆數]
    df.columns = df.columns.fillna('unnamed')
    columns = df.columns
    for c in df.columns:
        try:
            pat = '^.*述|借貸$'
            if re.match(pat, c):
                continue
        except TypeError:
            logger.error(f'欄名{c}非字串格式')
            continue

        pat = '^.*(金額|專戶|淨股利|次數|損益|[毛淨營]利|累計|差異|評分|期末|負債|營收|年數|\(元\))|成本|支出|存入|現值|借券|餘額|借|貸$'
        if re.match(pat, c):
            try:
                if (np.issubclass_(df[c].dtype.type, np.integer)  or
                    df[c].dtype == float or df[c].dtype == 'float64'
                ) and not c in 隱藏欄位:
                    try:
                        整數欄位.append(c)
                    except AttributeError:
                        整數欄位 = [c]
            except:
                breakpoint()
                pass
        pat = '^現金轉換天數|估計每股配發現金|估計股利|(元/股)|股價|配息|配股|r方|每股盈餘|.*符合度|.*比|.*指數|每股.*$'
        if re.match(pat, c):
            if df[c].dtype in [float] and not c in 隱藏欄位 and not c in 百分比欄位 and not c in 整數欄位: 
                try:
                    實數欄位.append(c)
                except AttributeError:
                    實數欄位 = [c]
        pat = '^.*(漲跌幅|率|比例|趨勢|\(%\))$'
        if re.match(pat, c):
            if df[c].dtype == float and not c in 隱藏欄位: 
                try:
                    百分比欄位.append(c)
                except AttributeError:
                    百分比欄位 = [c]
        pat = '.*(日期|時期).*'
        if re.match(pat, c):
            if df[c].dtype == "datetime64[ns]" and not c in 隱藏欄位:
                try:
                    日期欄位.append(c)
                except AttributeError:
                    日期欄位 = [c]
    if 整數欄位: 整數欄位 = [*set(整數欄位)]
    if 實數欄位: 實數欄位 = [*set(實數欄位)]
    if 百分比欄位: 百分比欄位 = [*set(百分比欄位)]
    from itertools import chain
    最大值顯著欄位.extend(整數欄位)
    最大值顯著欄位.extend(實數欄位)
    最大值顯著欄位.extend(百分比欄位)
    最大值顯著欄位 = [*set(最大值顯著欄位)]
    最大值顯著欄位 = [x for x in 最大值顯著欄位 
                         if x not in 百分比漸層欄位 
                         or x not in 不排序欄位]
    s = df.style.pipe(標準格式(整數欄位, 實數欄位, 百分比欄位
                              ,日期欄位
                              ,最大值顯著欄位
                              ,隱藏欄位
                              ,百分比漸層欄位
                              ,採用民國日期格式=採用民國日期格式
                              ))
    tp = df.copy()
    for c in df.columns: tp.loc[c] = c

    if 除錯提示:
        from pathlib import Path
        html = Path.home() / 'TEMP' / 'output.html'
        tp.to_html(html)
        import os
        os.system(f'start {html}')

    if 顯示提示:
        s = s.set_tooltips(tp)

    if 顯示:
        from pathlib import Path
        html = Path.home() / 'TEMP' / 'showdf.html'
        html.parent.mkdir(parents=True, exist_ok=True)
        s.to_html(html, encoding='utf8', doctype_html=True)
        import os
        os.system(f'start {html}')
    return s

def read_csv(*args, **kwargs):
    '編碼錯誤預設使用 replace。'
    from pathlib import Path
    fn = args[0]
    if isinstance(fn, Path):
        fn = str(fn)
    with open(fn, encoding=kwargs['encoding'], errors='replace') as f:
        import pandas as pd
        df = pd.read_csv(f, *args[1:], **kwargs)
        return df

def read_fwf(*args, **kwargs): 
    import pandas as pd
    mbyte_columns = kwargs['mbyte_columns']
    names = kwargs['names']
    encoding = kwargs['encoding']
    del kwargs['mbyte_columns']
    del kwargs['encoding']
    df = pd.read_fwf(args[0], header=None
                    ,converters={h:str for h in names}
                    ,encoding = 'cp437' # 利用cp437剛好用1位元編碼特性，表達ByteString
                    ,**kwargs
                    )
    def convert(cp437:str):
        try:
            return cp437.encode('cp437').decode(encoding, errors='replace') 
        except AttributeError:
            return cp437
    for c in mbyte_columns: 
        df[c] = df[c].map(convert)
    return df

def read_docx_tables(filename, tab_id=None, **kwargs):
    from warnings import warn
    warn("警告：【read_docx_tables】函數將廢棄，請改用【read_docx】函數！")
    return read_docx_tables(filename, tab_id, **kwargs)

def read_docx(filename, tab_id=None, **kwargs):
    """
    讀取 Word 文件(.docx)表格至 Pnadas DataFrame
    parse table(s) from a Word Document (.docx) into Pandas DataFrame(s)

    Parameters:
        filename:   file name of a Word Document

        tab_id:     parse a single table with the index: [tab_id] (counting from 0).
                    When [None] - return a list of DataFrames (parse all tables)

        kwargs:     arguments to pass to `pd.read_csv()` function

    Return: a single DataFrame if tab_id != None or a list of DataFrames otherwise
    """
    from docx import Document
    def read_docx_tab(tab, **kwargs):
        import io
        import csv
        vf = io.StringIO()
        writer = csv.writer(vf)
        for row in tab.rows:
            writer.writerow(cell.text for cell in row.cells)
        vf.seek(0)
        import pandas as pd
        return pd.read_csv(vf, **kwargs)

    doc = Document(filename)
    if tab_id is None:
        return [read_docx_tab(tab, **kwargs) for tab in doc.tables]
    else:
        try:
            return read_docx_tab(doc.tables[tab_id], **kwargs)
        except IndexError:
            print('Error: specified [tab_id]: {}  does not exist.'.format(tab_id))
            raise

def 分割鏈(鏈, 長度):
    from itertools import islice
    it = iter(鏈)
    size = 長度
    chunk = list(iter(lambda: list(islice(it, size)), []))
    return chunk

def 製作排行榜(排行個數或名次清單, 排行鍵, 漸增=False):
    def _製作排行榜(資料函數):
        from functools import wraps
        @wraps(資料函數)
        def 排行榜(*args, **kargs):
            df = 資料函數(*args, **kargs)
            df.sort_values(排行鍵, ascending=漸增, inplace=True)
            df.reset_index(drop=True, inplace=True)
            df.index = df.index+1
            df.index.name = '名次'
            if isinstance(排行個數或名次清單, int):
                排行個數 = 排行個數或名次清單 
                df = df.iloc[:排行個數]
            elif isinstance(排行個數或名次清單, list):
                名次清單 = 排行個數或名次清單 
                df = df.iloc[map(lambda n: n-1, 名次清單)]
            return df
        return 排行榜
    return _製作排行榜

def 連續增減次數(時序資料):
    資料名 = 時序資料.name
    df = 時序資料.to_frame()
    df['index']= df.index
    當期值 = df[資料名] 
    前期值 = 當期值.shift()
    環比 = 當期值 - 前期值
    消長 = 環比/環比.abs()
    變化 = 消長 != 消長.shift()
    df['分組']= 變化.cumsum()
    連續次數 = df.groupby('分組')['index'].rank(method='first')
    連續次數 = 連續次數 * 消長
    return 連續次數 

def 清查資料表非全數值欄(df, 列出非數值明細之欄=None):
    import pandas as pd

    # 將所有值轉換為數字類型，無法轉換的值設置為 NaN
    df_converted = df.apply(pd.to_numeric, errors='coerce')

    # 找出無法轉換的值
    invalid_cells = df_converted.isna()

    # 遍歷所有無法轉換的值並輸出
    for column in df.columns:
        if invalid_cells[column].any():
            invalid_values = df[column][invalid_cells[column]]
            logger.warn(f"{column} 欄有無法轉換成數值之資料值！")
            if column in 列出非數值明細之欄:
                logger.warn(f"非數值之資料值及其索引: {invalid_values.index.tolist()} - {invalid_values.tolist()}")
            else:
                logger.debug(f"非數值之資料值及其索引: {invalid_values.index.tolist()} - {invalid_values.tolist()}")

def 重名加序(columns):
    seen = {}
    new_columns = []
    for col in columns:
        if col in seen:
            seen[col] += 1
            new_columns.append(f"{col}.{seen[col]}")
        else:
            seen[col] = 0
            new_columns.append(col)
    return new_columns




def 修正英特爾網卡無線網路下載速度緩慢問題():
    '參考：https://www.reddit.com/r/HomeNetworking/comments/14slwlo/slow_wifi_speeds_with_ax211_intel_wifi_card/'
    import os
    # cmd = 'netsh interface set interface "Wi-Fi" admin=disable'
    # os.system(cmd)
    # cmd = 'netsh interface set interface "Wi-Fi" admin=enable'
    # os.system(cmd)
    cmd = 'netsh int tcp set global autotuninglevel=normal'
    os.system(cmd)

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-w", "--fix_intel_wifi_slow"
                       ,help="修正英特爾網卡無線網路下載速度緩慢問題"
                       ,action="store_true")
    args = parser.parse_args()
    if args.fix_intel_wifi_slow:
        修正英特爾網卡無線網路下載速度緩慢問題()



import unittest
from pathlib import Path
wdir = Path(__file__).parent
class Test(unittest.TestCase):

    def test_轉數值(self):
        from zhongwen.number import 中文數字轉數值, 轉數值, 約數, 百分比
        self.assertEqual(轉數值('十一萬'), 110000)
        self.assertEqual(轉數值('一五０'), 150)
        self.assertEqual(轉數值('一五0'), 150)
        self.assertEqual(轉數值('二五'), 25)
        self.assertEqual(轉數值('一００'), 100)
        self.assertEqual(轉數值('- 6.35'), -6.35)
        self.assertEqual(轉數值('35,636.71'), 35636.71)
        self.assertEqual(轉數值('一萬三千二百五十九'), 13259)
        self.assertEqual(轉數值('參佰玖拾柒萬肆仟壹拾伍'), 3974015)
        self.assertEqual(轉數值('5,000'), 5000)
        self.assertEqual(轉數值('8.1%'), 0.081)
        self.assertEqual(轉數值('-8.1%'), -0.081)
        self.assertEqual(轉數值('48.89%'), 0.4889)
        self.assertEqual(轉數值('１２３'), 123)
        self.assertEqual(轉數值('3年'), 3)
        self.assertEqual(約數('55,302'), '5萬餘')
        self.assertEqual(約數('5,302'), '5,302')
        self.assertEqual(百分比('0.8911'), '89.11％')

        self.assertEqual(中文數字轉數值('一'), 1)
        self.assertEqual(中文數字轉數值('二十五'), 25)
        self.assertEqual(中文數字轉數值('一萬三千二百五十九'), 13259)
        self.assertEqual(中文數字轉數值('十一'), 11)
        self.assertEqual(中文數字轉數值('拾柒'), 17)
        self.assertEqual(中文數字轉數值('十萬'), 100000)
        self.assertEqual(中文數字轉數值('十一萬'), 110000)

    def test_中文數字(self):
        from zhongwen.number import 中文數字, 中文数字, 大寫中文數字
        self.assertEqual(中文數字(16), '十六')
        self.assertEqual(中文數字(110), '一百一')
        self.assertEqual(中文數字(1600), '一千六')
        self.assertEqual(中文数字(3.4), '三点四')
        self.assertEqual(中文数字(10600), '一万零六百')
        self.assertEqual(中文数字(1821010), '一百八十二万一千零一十')
        self.assertEqual(中文数字(111180000), '一亿一千一百一十八万')
        self.assertEqual(大寫中文數字(21), '貳拾壹')
        self.assertEqual(大寫中文數字(23232.00518)
                        ,'貳萬參仟貳佰參拾貳點零零伍壹捌')
        self.assertEqual(中文數字(23232.00518, 兩=True)
                        ,'兩萬三千兩百三十兩點零零五一八')

        from zhongwen.number import 標號, 轉標號
        self.assertEqual(str(標號(1, 1)), '壹、' )
        self.assertEqual(str(標號(1, 4)), '(一)')
        self.assertEqual(str(標號(1, 5)), '1.')
        self.assertEqual(轉標號('壹、'), 標號(1, 1))
        self.assertEqual(轉標號('貳拾、'), 標號(20, 1))
        self.assertEqual(轉標號('一、'), 標號(1, 3))
        self.assertEqual(轉標號('3.'), 標號(3, 5))
        self.assertEqual(轉標號('1.'), 標號(1, 5))

        from zhongwen.number import 全型數字
        self.assertEqual(全型數字('123'), '１２３')

    def test_counter_sec_ones(self):
        from zhongwen.number import 連續正負次數
        a = [2, 3, -0.1, 7, 8, 9, -0.3, 2, 3, 4, 5, 5]
        self.assertEqual(連續正負次數(a), 2)
        a = [-2, -3, -0.1, 7, 8, 9, -0.3, 2, 3, 4, 5, 5]
        self.assertEqual(連續正負次數(a), -3)
        a = [-2, -3, -0.1, -7]
        self.assertEqual(連續正負次數(a), -4)
        a = [2, 3, 0.1, 7]
        self.assertEqual(連續正負次數(a), 4)

        from zhongwen.number import 連續正數
        a = [2, 3, -0.1, 7, 8, 9, -0.3, 2, 3, 4, 5, 5]
        self.assertEqual(連續正數(a), 5)
        a = [2]
        self.assertEqual(連續正數(a), 1)
        a = [2, 3, -0.1, 7, 8, 9, -0.3, 3, 4, 5, 5]
        self.assertEqual(連續正數(a), 4)

        # from zhongwen.number import 最近連續正負次數
        # a = [2]
        # self.assertEqual(最近連續正負次數(a), 1)
        # a = [-1]
        # self.assertEqual(最近連續正負次數(a), -1)
        # a = [2, 7]
        # self.assertEqual(最近連續正負次數(a), 2)
        # a = [2, -7]
        # self.assertEqual(最近連續正負次數(a), -1)
        # a = [float('nan'), 3, 0.1, 7]
        # self.assertEqual(最近連續正負次數(a), 3)

    def test_nan_handler(self):
        from zhongwen.number import 除零例外則回覆非數常數
        div_zero_func = lambda r: r/0
        div_zero_return_nan_func = 除零例外則回覆非數常數(div_zero_func)
        import numpy as np
        self.assertTrue(np.isnan(div_zero_return_nan_func(1)))

    def test_文字表達(self):
        from zhongwen.number import 增減百分比, 增減
        self.assertEqual(增減百分比(0.11), '增加11.00%')
        self.assertEqual(增減百分比(-0.01), '減少1.00%')
        self.assertEqual(增減(3), '增加3')
        self.assertEqual(增減(-2), '減少2')

    def test_script(self):
        from subprocess import check_output
        out = check_output("py -m zhongwen.number --increment 貳拾、"
                          ,shell=True)
        self.assertEqual(out.decode('cp950').rstrip(), '貳拾壹、')

        out = check_output("py -m zhongwen.number --increment 1."
                          ,shell=True)
        self.assertEqual(out.decode('cp950').rstrip(), '2.')
      
if __name__ == '__main__':
    unittest.main()



'日期處理'
from datetime import date
from functools import lru_cache
from pandas import Timestamp, Timedelta

def 全是日期嗎(vs):
    from datetime import datetime
    import pandas as pd
    from collections.abc import Iterable 
    match vs:
        case list():
            return all(全是日期嗎(v) for v in vs)
        case pd.Series():
            return vs.map(全是日期嗎).all()
        case _:
            v = vs
            return not pd.isnull(v) and (isinstance(v, pd.Timestamp) or
                                         isinstance(v, datetime) or
                                         isinstance(v, date) 
                                         )

def 有日期嗎(vs):
    from datetime import datetime
    import pandas as pd
    from collections.abc import Iterable 
    match vs:
        case list():
            return any(全是日期嗎(v) for v in vs)
        case pd.Series():
            return vs.map(全是日期嗎).any()
        case _:
            v = vs
            return not pd.isnull(v) and (isinstance(v, pd.Timestamp) or
                                         isinstance(v, datetime) or
                                         isinstance(v, date) 
                                         )



def 取過去日期(d):
    return 取日期(d, 日期大於今日省略年推論為去年=True)

def 昨日():
    return 今日() - Timedelta(days=1)

def 取日期(d=None, 錯誤為空值=True, first=True, defaulttoday=True, default=None,
           日期大於今日省略年推論為去年=False) -> Timestamp:
    '日期表示 d 如為 None，傳回今日。'
    import re
    import pandas as pd
    from datetime import datetime, timedelta, date
    if not default: 
        default=pd.NaT
    if pd.isnull(d):
        return pd.NaT
    match d:
        case float():
            return 取日期(f'{d:.0f}')
        case int():
            if d > 2_01_01:
                return 取日期(f'{d:07}')
            return 取日期(f'{d:05}')
        case datetime():
            return Timestamp(d).normalize()
        case date():
            return Timestamp(d).normalize()
        case Timestamp():
            return Timestamp(d).normalize()
        case pd.Period():
            return d.end_time
        case str(d):
            pat = r'^昨日?$'
            if m:=re.match(pat, d):
                return 昨日()
            # 省略日自動推論為月底
            pat = r'^(\d\d\d)(\d\d)$'
            if m:=re.match(pat, d):
                year = int(m[1]) + 1911
                mon = int(m[2])
                import calendar
                _, last_day = calendar.monthrange(year, mon)
                return Timestamp(year, mon, last_day).normalize()
            # 省略年自動推論為今年
            pat1 = r'(\d{1,2})([./])\d{1,2}'
            pat2 = r'^(\d\d)()(\d\d)$' # 中間空白括號是表達日期分隔為空字元
            if m:=(re.match(pat1, d) or re.match(pat2, d)):
                if int(m[1]) <= 12: 
                    year = Timestamp.now().year
                    od = d
                    sep = m[2]
                    d = 取日期(f'{year}{sep}{d}')
                    if 日期大於今日省略年推論為去年 and d > 取日期(datetime.now()): 
                        # 省略年推論為今年大於今日，則推論為去年
                        return 取日期(f'{year-1}{sep}{od}')
                    return d
         
            # 日期2022/5/27
            pat = r'\d{4}([-./])\d{1,2}[-./]\d{1,2}'
            if m:=re.match(pat, d):
                s = m[1]
                try:
                    return Timestamp(datetime.strptime(m[0], f'%Y{s}%m{s}%d')).normalize()
                except:breakpoint()

            # 日期20220527
            pat = r'\d{4}\d{2}\d{2}'
            if m:=re.match(pat, d):
                return Timestamp(datetime.strptime(m[0], f'%Y%m%d')).normalize()
 
            # 民國日期帶增減日數形式，如【111.4.29+150】。
            from pandas import Timedelta
            pat = r'(\d{2,3})[-./](\d{1,2})[-./](\d{1,2})([+])(\d+)'
            if m:=re.match(pat, d):
                return Timestamp(int(m[1])+1911, int(m[2]), int(m[3])
                                 ).normalize()+Timedelta(days=int(m[5]))

            # 民國日期形式如980731, 1110527。
            pat = r'(\d{2,3})(\d{2})(\d{2})'
            if m:=re.match(pat, d):
                try:
                    return Timestamp(int(m[1])+1911, int(m[2]), int(m[3])).normalize()
                except ValueError as e: 
                    if 錯誤為空值: return pd.NaT
                    raise ValueError(f'【{d}】引發例外{e}！')

            # 民國日期形式如980731, 1110527。
            pat = r'(\d{2,3})(\d{2})(\d{2})'

            # 民國日期其形式如 109/05/29 、109.05.29及109-5-29 等。
            pat = r'(\d{2,3})[-./](\d{1,2})[-./](\d{1,2})'
            if m:=re.match(pat, d):
                return Timestamp(int(m[1])+1911, int(m[2]), int(m[3])).normalize()

            pat = r'民國\s*(\d+)\s*年\s*(\d+)\s*月\s*(\d+)\s*日'
            if m:=re.match(pat, d):
                try:
                    return Timestamp(int(m[1])+1911, int(m[2]), int(m[3])).normalize()
                except ValueError: return pd.NaT
            pat = r'(\d+)年\D*(\d+)月'
            if m:=re.match(pat, d):
                try:
                    return 月底(Timestamp(int(m[1])+1911, int(m[2]), 1))
                except ValueError: return pd.NaT
            pat = r"'\d\d/\d\d/\d\d"
            if m:=re.match(pat, d):
                try:
                    return 取日期(m[0].replace("'", "20"))
                except ValueError: return pd.NaT
 
            if default: return default 
            return d
        case _:
            if defaulttoday:
                return Timestamp.today()
            raise TypeError(f'不支援類型[{type(d)}]、值[{d}]！')

def 取日數(期間):
    try:
        return 期間.days
    except:
        return float('nan')

def 上月() -> Timestamp:
    '上月底'
    from datetime import timedelta
    return 今日().replace(day=1) - timedelta(days=1)

def 上上月() -> Timestamp:
    '上上月底'
    from datetime import timedelta
    return 上月().replace(day=1) - timedelta(days=1)

def 民國日期(d=None, fmt='%Y%m%d', 昨今明表達=False):
    '%Y表年數、%m表月數前置0、%d表日數前置0、%M表月數不前置0'
    if not d: d = 今日()
    d = 取日期(d)
    import pandas as pd
    if pd.isnull(d):
        return ''
    if 昨今明表達:
        if d == 今日()-Timedelta(days=1):
            return '昨'
        if d == 今日():
            return '今'
        if d == 今日()+Timedelta(days=1):
            return '明'
        if d == 今日()+Timedelta(days=2):
            return '後天'

    fmt = fmt.replace(
            '%Y', '%(year)03d'
            ).replace(
            '%m', '%(month)02d'
            ).replace(
            '%M', '%(month)d'
            ).replace(
            '%d', '%(date)02d'
            ).replace(
            '%D', '%(date)d'
            )

    year = d.year-1911
    return fmt % {"year":year, "month":d.month, "date":d.day}

def 民國正式日期(d=None):
    '格式如：112年7月29日'
    if not d:
        d = 今日()
    return 民國日期(d, "%Y年%M月%D日")

def 民國月份(d):
    '格式如：112年7月'
    return 民國日期(d, "%Y年%M月")

def 經過日數(起, 迄):
    起=取日期(起)
    迄=取日期(迄)
    return (迄-起).days

def 約年(日數):
    return f'{日數/365:0.0f}年餘'

def 逾(起, 迄):
    return f'逾{約年(r(end) - r(start))}'

def 上年度():
    d = date.today()
    return f'{d.year-1911-1}年度'

def 本年度():
    d = date.today()
    return f'{d.year-1911}年度'

def 上上年度():
    d = date.today()
    return f'{d.year-1911-2}年度'

def 今日() -> Timestamp:
    return Timestamp.today().normalize()

def 前天() -> Timestamp:
    return 今日() + Timedelta(days=-2)

def 明天() -> Timestamp:
    return 今日() + Timedelta(days=1)

def 近三日() -> Timestamp:
    curdate = 前天()
    while curdate <= 今日():
        yield curdate
        curdate += Timedelta(days=1)

def 上年() -> Timestamp:
    d = 今日()
    return d.replace(year=d.year-1)

def 上年初() -> Timestamp:
    d = 今日()
    return d.replace(year=d.year-1, month=1, day=1)

def 去年底() -> Timestamp:
    d = 今日()
    return d.replace(year=d.year-1, month=12, day=31)

def 上年底() -> Timestamp:
    return 去年底()

def 前年底() -> Timestamp:
    d = 今日()
    return d.replace(year=d.year-2, month=12, day=31)

def 年初(年數=None) -> Timestamp:
    if 年數:
        return Timestamp(年數, 1, 1)
    d = 今日()
    return d.replace(month=1, day=1)

def 年底(年數=None) -> Timestamp:
    if 年數:
        return Timestamp(年數, 12, 31)
    d = 今日()
    return d.replace(month=12, day=31)

def 民國年底(民國年數) -> Timestamp:
    if 民國年數:
        return Timestamp(民國年數+1911, 12, 31)
    d = 今日()
    return d.replace(month=12, day=31)

def 季別(日期=今日()) -> (int, int):
    '指定日期之季數，以(年數, 季數)表達，未指定為今日歸屬季別'
    d = 日期
    if 1 <= d.month <= 3: return (d.year, 1)
    if 4 <= d.month <= 6: return (d.year, 2)
    if 7 <= d.month <= 9: return (d.year, 3)
    if 10 <= d.month <= 12: return (d.year, 4)

def 季數(日期=今日()):
    from warnings import warn
    warn(f'配合證交所資料項目名稱，【季數】將廢棄，請使用【季別】', DeprecationWarning, stacklevel=2)
    return 季別(日期)

def 季初(年數或日期=None, 季別參數=None) -> Timestamp:
    '指定季之最初日，未指定為本季'
    y, q = 季別()
    if 全是日期嗎(年數或日期):
        d = 年數或日期
        y, q = 季別(d)
    if 年數或日期 and 季別參數:
        y = 年數或日期 
        q = 季別參數
    match q:
        case 1: return Timestamp(y,  1, 1)
        case 2: return Timestamp(y,  4, 1)
        case 3: return Timestamp(y,  7, 1)
        case 4: return Timestamp(y, 10, 1)
    raise ValueError(f'年數或日期值「{年數或日期}」及季別參數值「{季別參數}」係錯誤值！')

def 季末(年數或日期=None, 季別參數=None):
    '指定季之最末日，未指定為本季末'
    y, q = 季別()
    if 全是日期嗎(年數或日期):
        d = 年數或日期
        y, q = 季別(d)
    if 年數或日期 and 季別參數:
        y = int(年數或日期 )
        q = int(季別參數)
    match q:
        case 1: return Timestamp(y, 3, 31)
        case 2: return Timestamp(y, 6, 30)
        case 3: return Timestamp(y, 9, 30)
        case 4: return Timestamp(y, 12, 31)

def 本季初():
    from warnings import warn
    warn(f'【季初】無參數即為本季初，原【本季初】函數將廢棄！', DeprecationWarning, stacklevel=2)
    return 季初() 

def 上季數() -> (int, int):
    '上季數之(年數, 季數)。'
    from datetime import timedelta
    return 季別(本季初()-timedelta(days=1))

def 上季初():
    return 季初(*上季數())

def 月底(日期或年數=None, 月數=None):
    '無指定參數則為本月底。'
    import calendar
    年數 = 日期或年數
    if not 日期或年數:
        年數=今日().year
    if not 月數:
        月數=今日().month
    if 全是日期嗎(日期或年數):
        年數=日期或年數.year
        月數=日期或年數.month
    月底日數 = calendar.monthrange(年數, 月數)[1]
    return Timestamp(年數, 月數, 月底日數)

def 月末(日期或年數=None, 月數=None):
    from warnings import warn
    warn(f'因【月末】名稱較罕用將廢棄，請改用【月底】', DeprecationWarning, stacklevel=2)
    return 月底(日期或年數, 月數)

def 與季末相距月數(日期):
    from dateutil.relativedelta import relativedelta
    return relativedelta(季末(日期), 日期).months

def 與年底相距月數(日期):
    from dateutil.relativedelta import relativedelta
    return relativedelta(年底(), 日期).months

def 月起迄(year, mon):
    月初 = Timestamp(year, mon, 1)
    return [月初, 月底(year, mon)]

def 前幾月(月數):
    from dateutil.relativedelta import relativedelta
    return 今日() - relativedelta(months=月數)

def 近幾個月底(月數):
    from datetime import timedelta
    import pandas as pd
    start = 前幾月(月數)
    end = 今日()
    curdate = 月末(start)
    while curdate <= end:
        yield curdate
        curdate += timedelta(days=1)
        curdate = 月末(curdate)

def 前幾季(季數):
    from dateutil.relativedelta import relativedelta
    return 季末(今日() - relativedelta(months=季數*3))

def 上季():
    from dateutil.relativedelta import relativedelta
    return 前幾季(1)

def 近幾季(季數):
    from datetime import timedelta
    start = 前幾季(季數)
    end = 今日()
    curdate = 季末(start)
    while curdate <= end:
        yield curdate
        curdate += timedelta(days=1)
        curdate = 季末(curdate)

def 迄每季(起季):
    '自起季迄每季'
    from datetime import timedelta
    end = 今日()
    curdate = 季末(起季)
    while curdate <= end:
        yield curdate
        curdate += timedelta(days=1)
        curdate = 季末(curdate)

def 民國年月(日期=None):
    '日期格式如112年8月'
    if not 日期: 日期 = 今日()
    return 民國日期(日期, '%Y年%M月')

def 民國年數(日期=None):
    '日期格式如112年'
    if not 日期: 日期 = 今日()
    return 日期.year - 1911

def 今民國年數():
    return 民國年數()

def 今年數():
    return 今日().year

def 民國年(日期=None):
    '取民國年表達字串，格式如112年。'
    return f'{民國年數(日期)}年'

def 民國季別(日期=None):
    '季別格式如112年第1季'
    if not 日期: 日期 = 今日()
    年數, 季數 = 季別(日期)
    return f'{年數-1911}年第{季數}季'

def 學期(日期=None):
    if not 日期:
        日期=今日()
    學年度 = 日期.year-1911
    if 1 < 日期.month < 7:
        學年度-=1
        return f'{學年度}年下'
    return f'{學年度}年上'

def 應公布財報季別(公司類型='一般公司') -> (int, int):
    '''傳回目前應公布財報年數及季別數組，
一般公司最近一次應公布財報年季。
證交所規定上市公司公告申報財務報表之期限如下：
一、年度財務報告：每會計年度終了後3個月內(3/31前)。
二、第一季、第二季、第三季財務報告：
(一)一般公司(含投控公司)：每會計年度第1季、第2季及第3季終了後45日內(5/15、8/14、11/14前)。
(二)保險公司：每會計年度第1季、第3季終了後1個月內(4/30、10/31前)，每會計年度第2季終了後2個月內(8/31前)。
(三)金控、證券、銀行及票劵公司：每會計年度第1季、第3季終了後45日內(5/15、11/14前)，每會計年度第2季終了後2個月內(8/31前)。惟金控公司編製第1季、第3季財務報告時，若作業時間確有不及，應於每季終了後60日內(5/30、11/29前)補正。'''
    日期 = 今日()
    年 = 今日().year
    q1 = Timestamp(年, 5, 15)
    q2 = Timestamp(年, 8, 14)
    q3 = Timestamp(年, 11, 14)
    q4 = Timestamp(年, 3, 31)
    if q4 <= 日期 < q1 :
        年 = 年-1
        季 = 4
    elif q1 < 日期 < q2:
        季 = 1
    elif q2 < 日期 <= q3:
        季 = 2
    elif q3 <= 今日():
        季 = 3
    elif 今日() < q4:
        年 = 年-1
        季 = 3
    return 年, 季 

def 自起算民國年逐年列舉迄今年(起算民國年數):
    '自起算民國年列舉至今年底'
    起算年數 = 民國年底(起算民國年數).year
    今年數 = 年底().year
    return [年底(y) for y in range(起算年數, 今年數+1)]

def 自起算民國年初逐月列舉迄上個月(起算民國年數):
    '民國年數'
    from datetime import timedelta
    start = 取日期(f'{起算民國年數}.1.1')
    end = 上月()
    curdate = 月末(start)
    while curdate <= end:
        yield curdate
        curdate += timedelta(days=1)
        curdate = 月末(curdate)

def 自起算民國年月逐月列舉迄上個月(起算民國年數, 起算月) -> Timestamp:
    '民國年數'
    start = 取日期(f'{起算民國年數}.{起算月}.1')
    end = 上月()
    curdate = 月末(start)
    while curdate <= end:
        yield curdate
        curdate += Timedelta(days=1)
        curdate = 月末(curdate)

def 自起日按日列舉迄今(起日):
    curdate = 起日
    while curdate <= 今日():
        yield curdate
        curdate += Timedelta(days=1)

def 自年初按日列舉迄今():
    yield from 自起日按日列舉迄今(年初())

def 相距年數(迄, 起):
    date1 = 起
    date2 = 迄
    years_diff_exact = date2.year - date1.year
    if (date2.month, date2.day) < (date1.month, date1.day):
        years_diff_exact -= 1
    return years_diff_exact 



import re
小寫數字表 = "零一二三四五六七八九"
大寫數字表 = "零壹貳參肆伍陸柒捌玖"
大写数字表 = "零壹贰叁肆伍陆柒捌玖"
小寫位名表 = "十百千"
大寫位名表 = "拾佰仟"
組名表 = "萬億兆京垓秭穰溝澗正載"
组名表 = "万亿兆京垓秭穰沟涧正载"
半型數字表 = "0123456789"
全型數字表 = "０１２３４５６７８９"

# 取得數值
def 轉數值(n, 傳回格式=False) -> int|float:  
    import pandas as pd
    if isinstance(n, str):
        # 百分比
        pat = r'(-?[\d]+(\.\d+)?)[%％]'
        if m:=re.match(pat, n):
            return 轉數值(m[1])/100
        pat = f'^[{全型數字表}]+$'
        if re.match(pat, n):
            d = str.maketrans(全型數字表, 半型數字表)
            return int(n.translate(d))

        # 不具位名之中文數
        中文數字='○0０零壹貳參肆伍陸柒捌玖一二三四五六七八九'
        pat = f'^[{中文數字}]+$'
        if re.match(pat, n):
            d = str.maketrans(中文數字, '0000123456789123456789')
            return int(n.translate(d))

        # 具單位之數值表示，如3年則取出3。
        單位 = '年'
        pat = rf'(\d+)[{單位}]+'
        if m:=re.match(pat, n):
            return int(m[1])

        # 具位名之中文數
        pat = r'^[零壹貳參肆伍陸柒捌玖拾一二三四五六七八九十百千佰仟萬億兆]+$'
        if m:=re.match(pat, n):
            return 中文數字轉數值(m.group(0))
        try:
            n = re.sub(r'[ ,，]', '', n)
            pat = r'^-?\d+$'
            if m:=re.match(pat, n):
                return int(n)
            return float(n)
        except:
            return 0


    return n

def 全型數字(number):
    if isinstance(number, int) or (isinstance(number, str) and number.isdigit()):
        translation_table = str.maketrans(半型數字表, 全型數字表)
        if isinstance(number, int):
            number_str = str(number)
        else:
            number_str = number
        return number_str.translate(translation_table)
    else:
        raise ValueError("Input must be a non-negative integer or a string of digits.")

def 中文數字轉數值(n):
    數字:str=小寫數字表+大寫數字表+大写数字表
    位名:str=小寫位名表+大寫位名表
    組名:str=組名表+组名表
    v = 0
    r = 0
    for c in n:
        if c in 數字:
            v = 數字.index(c) % len(小寫數字表)
        if c in 位名:
            if v != 0:
                r += v*10**(位名.index(c) % len(小寫位名表) + 1)
                v = 0
            if r == 0: # 十一 -> 11 
                r += 1*10**(位名.index(c) % len(小寫位名表) + 1)
        if c in 組名:
            r += v
            r *= 10**(4*(組名.index(c) % len(組名表) + 1))
            v = 0
    if v!=0:
        r+=v
    return r
    

def 中文數字(
    n: int | float | str,
    大寫: bool = False,
    簡體: bool = False,
    異體零: bool = False,
    兩: bool = False
) -> str:
    位名表 = 大寫位名表 if 大寫 else 小寫位名表
    _組名表 = 组名表 if 簡體 else 組名表
    數字表 = 大寫數字表 if 大寫 else 小寫數字表 
    數字表 = 大写数字表 if 大寫 and 簡體 else 數字表 
    數字表 = re.sub('二', '兩', 數字表) if 兩 else 數字表 
    點 =  '点' if 簡體 else '點' 
    
    n:str = str(n)
    i:str = n
    d = ""

    if "." in n:
        i, d= n.split(".", 1)
        d = d.rstrip("0")
    cn = ''
    i = list(reversed(i))

    def 轉中文數字(i):
        cn = ''
        for pos, digit in enumerate(i):
            位名 = ""
            if digit != '0' and pos > 0:
                位名 = 位名表[pos-1]
            cn = 數字表[int(digit)] + 位名 + cn
        return cn

    for pos, 一組阿拉伯數字 in enumerate([i[idx:idx + 4] for idx in range(0, len(i), 4)]):
        _cn = 轉中文數字(一組阿拉伯數字)
        if pos > 0: 
            _cn = _cn + _組名表[pos-1]
        cn = _cn + cn 
    #一萬零六百零零 -> 一萬零六百
    cn = re.sub('零+$', '', cn)
    #一十六 -> 十六
    cn = re.sub('^一十', '十', cn)
    #一千六百 -> 一千六
    cn = re.sub(f'(千[{數字表}])百$', r'\1', cn)
    #一百一十 -> 一百一
    cn = re.sub(f'(百[{數字表}])十$', r'\1', cn)
    if d != "":
        cn += 點 + d.translate(str.maketrans('0123456789', 數字表))
    return cn    

def 中文数字(n):
    return 中文數字(n, 簡體=True)

def 大寫中文數字(n):
    return 中文數字(n, 大寫=True)

def 大写中文数字(n):
    return 中文數字(n, 大寫=True, 簡體=True)

def 增減(變動數, 增減用語=['增加', '減少']):
    return (增減用語[0] if 變動數 > 0 else 增減用語[1]) + f'{abs(變動數):,.0f}'

# def 衰長(變動數):
    # return ('成長' if 變動數 > 0 else '減少') + f'{abs(變動數):,.0f}'

def 增減數(變動數):
    return ('增加' if 變動數 > 0 else '減少') + f'{abs(變動數)*100:,.2f}'

def 增減百分點(變動數, 增減用語=['增加', '減少']):
    return (增減用語[0] if 變動數 > 0 else 增減用語[1]) + f'{abs(變動數)*100:,.2f}'

def 約數(n) -> str:
    '中文萬約數表達，未達萬以全數表達'
    n = 轉數值(n)
    n = abs(n)
    if(n<10**4): return f'{n:,.0f}'
    s=""
    #億位數字
    h=n//10**8
    r=n%10**8
    if h:s+=f"{h:,.0f}億"
    #取萬位數字
    h=r//10**4
    r=n%10**4
    if h:s+=f"{h:,.0f}萬"
    if r: s+= "餘" 
    return s

def 最簡約數(n) -> str:
    '僅表達最高位數字，舉如：17億500萬餘元表達為17億餘元。'
    on = 轉數值(n)
    n = abs(on)
    sign = on / n if n > 0 else 1
    if(n<10**4): return f'{n:,.0f}'
    s=""
    #億位數字
    h=n//10**8
    r=n%10**8
    if h:
        s+=f"{h*sign:,.0f}億"
        if r: s+= "餘" 
        return s
    #萬位數字
    h=r//10**4
    r=n%10**4
    if h:s+=f"{h*sign:,.0f}萬"
    if r: s+= "餘" 
    return s

def 增減約數(變動數):
    return ('增加' if 變動數 > 0 else '減少') + 約數(abs(變動數))

def 百分比(n):
  '表達至2位小數百分比，另以--表達缺失值'
  import pandas as pd
  if pd.isnull(n): return '--'
  return f'{abs(float(n))*100:,.2f}％'

def 增減百分比(r, 增減文字=["增加", "減少"]):
    import pandas as pd
    if pd.isnull(r): return ''
    if isinstance(r, str): return r
    n = abs(r)
    _增減文字 = 增減文字[0] if r>0 else  增減文字[1] 
    return f'{_增減文字}{n:.2%}'

def 利損百分比(r):
    '用於利損項目，如毛利或毛損'
    return 增減百分比(r, ['利', '損'])

def 消長百分比(r):
    '比較前時序之數值用消長'
    return 增減百分比(r, ['成長', '衰退'])

def 同比百分比(r):
    '數據較去年同期增減'
    return 增減百分比(r, ['年增', '年減'])

def 二位實數(f):
    import pandas as pd
    if pd.isnull(f): return '--'
    return f'{abs(float(f)):,.2f}'

class 標號:
    def __init__(self, 號碼, 階層):
       self.號碼 = 號碼
       self.階層 = 階層

    def __str__(self):
        # 標號階層：壹、(貳)三、(四)5.(6)
        if self.階層==1:
            return f'{大寫中文數字(self.號碼)}、'
        if self.階層==2:
            return f'({大寫中文數字(self.號碼)})'
        if self.階層==3:
            return f'{中文數字(self.號碼)}、'
        if self.階層==4:
            return f'({中文數字(self.號碼)})'
        if self.階層==5:
            return f'{self.號碼}.'
        if self.階層==6:
            return f'({self.號碼})'

    def __eq__(self, other):
        if isinstance(other, 標號):
            return self.號碼==other.號碼 and self.階層==other.階層

def 轉標號(text):
    pat = f'([{大寫數字表}{大寫位名表}{組名表}]+)、'
    if m:=re.match(pat, text):
        return 標號(轉數值(m[1]), 1)
    pat = f'\\(([{大寫數字表}{大寫位名表}{組名表}]+)\\)'
    if m:=re.match(pat, text):
        return 標號(轉數值(m[1]), 2)
    pat = f'([{小寫數字表}{小寫位名表}{組名表}]+)、'
    if m:=re.match(pat, text):
        return 標號(轉數值(m[1]), 3)
    pat = f'\\(([{小寫數字表}{小寫位名表}{組名表}]+)\\)'
    if m:=re.match(pat, text):
        return 標號(轉數值(m[1]), 4)
    pat = f'([\\d]+).'
    if m:=re.match(pat, text):
        return 標號(轉數值(m[1]), 5)
    return text 
    pat = f'\\(([\\d+)\\)'
    if m:=re.match(pat, text):
        return 標號(轉數值(m[1]), 5)
    return text 
 
def 連續正負次數(數列):
    import numpy as np
    a = np.sign(數列)
    s = 0
    for n in a:
        if s*n < 0: return s
        s+=n
    return s

def 最近連續正負次數(時序):
    from warnings import warn
    from zhongwen.時序分析 import 最近連續正負次數
    warn(f'【最近連續正負次數】已移到「時序分析」模組，請使用【from 時序分析 import 最近連續正負次數】。', DeprecationWarning, stacklevel=2)
    return 最近連續正負次數(時序)

def 連續正數(數列):
    import numpy as np
    n = np.array(數列)
    n = n > 0
    s = 0
    for v in n:
        s = v*(s+v) 
    return s

def 貸方科目收支金額轉借方金額(金額):
    return abs(min(0, 金額))

def 貸方科目收支金額轉貸方金額(金額):
    return abs(max(0, 金額))

def 除零例外則回覆非數常數(含除法函數):
    '非數常數即NaN，亦即IEEE754 之 Not a Number 浮點數值。'
    from functools import wraps
    @wraps(含除法函數)
    def 除零回覆非數常數(*args, **kargs):
        try:
            return 含除法函數(*args, **kargs)
        except (ZeroDivisionError, TypeError):
            return float('nan')
    return 除零回覆非數常數

def 發生例外則回覆非數常數(函數):
    '非數常數即NaN，亦即IEEE754 之 Not a Number 浮點數值。'
    from functools import wraps
    @wraps(函數)
    def 發生例外回覆非數常數(*args, **kargs):
        try:
            return 函數(*args, **kargs)
        except:
            return float('nan')
    return 發生例外回覆非數常數

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--int", help="整數")
    parser.add_argument("--about", help="約數")
    parser.add_argument("--float", help="小數")
    parser.add_argument("--percent", help="百分比")
    parser.add_argument("--test", help="測試", action='store_true')
    parser.add_argument("--increment", help="中文數遞增")
    parser.add_argument("--decrement", help="中文數遞減")
    parser.add_argument("--dec_level", help="中文標號遞減層級")
    args = parser.parse_args()
    if args.increment:
        n = 轉標號(args.increment)
        # breakpoint()
        print(標號(n.號碼+1, n.階層))
    elif args.decrement:
        print(中文數遞減(args.decrement))
    elif args.percent:
        print(百分比(轉數值(args.percent)))
    elif args.float:
        print(小數(轉數值(args.float)))
    elif args.about:
        print(約數(轉數值(args.about)))
    elif args.int:
        print(f'{轉數值(args.int):,}')
    elif args.dec_level:
        n, l = 標號(args.dec_level)
        print(中文標號(n, l-1))
    elif args.test:
        test()



import pandas as pd
import re

def 中央地方法規條文():
    from zhongwen import 中央法規
    from zhongwen import 花蓮縣法規
    df = pd.concat([中央法規.法規條文(), 花蓮縣法規.法規條文()])
    df['條號'] = df.條號.map(lambda n: f'{n}')
    return df

def 顯示關鍵字查詢法規結果(關鍵字):
    from zhongwen.pandas_tools import 強調關鍵字, show_html
    df = 中央地方法規條文()
    pat = '|'.join(關鍵字)
    df = df.query('法規名稱.str.contains(@pat, na=False) or 條文內容.str.contains(@pat, na=False)')
    df = 強調關鍵字(df, ['法規名稱', '條文內容'], 關鍵字)
    df.reset_index(drop=True, inplace=True)
    df.index = df.index+1
    df.index.name = '編號'
    show_html(df, 無格式=False, 顯示筆數=300)

def 法條查詢(s):
    q = LawQuery(s)
    if not q.法規名稱 : return ''

    def 屬性條件(a):
        if a=='關鍵字': return '條文內容.str.contains(@q.關鍵字)'
        return f'{a}==@q.{a}'

    qstr = [屬性條件(a) for a in dir(q) if not a.startswith("__") and getattr(q, a)!=None]

    qstr = ' and '.join(qstr)
    s = 中央地方法規條文().query(qstr)
    return s

def 法條展開(s):
    return '\n'.join([f'{l[0]}第{l[2]}條規定：「{l[3]}」' for l in 法條查詢(s).values])

def 法條說明(s):
    ls = 法條查詢(s)
    def 條文內容(l):
        return f'{l[0]}第{l[2]}條規定：「{l[3]}」'
    doc = ''.join([條文內容(l.tolist()) for l in ls.values])
    return doc

法規類別 = r'(法|條例|規則|標準|細則|措施|準則)'

class LawQuery:
    def __init__(self, s):
        self.法規名稱=None
        self.條號=None
        self.關鍵字=None
        pat = rf'(.*{法規類別})第([-\d]+)[點條]'
        if m:=re.match(pat, s):
            self.法規名稱=m[1]
            self.條號 = m[3]
            return

        pat = rf'(.*{法規類別})([-\d]+)'
        if m:=re.match(pat, s):
            self.法規名稱=m[1]
            self.條號 = m[3]
            return

        pat = rf'(.*{法規類別})\[(.*)\]'
        if m:=re.match(pat, s):
            self.法規名稱 = m[1]
            self.關鍵字 = m[3]
            return

def 法規名稱字首樹():
    中央地方法規條文() 
    from marisa_trie import Trie
    return Trie(中央地方法規條文().法規名稱.to_list())

def 法規自動完成建議(line):
    '第一個結果是找出的字首'
    for i in range(0, len(line)):
        prefix = line[i:]
        ls = 法規名稱字首樹().keys(prefix)
        if len(ls) > 0:
            # 名稱越短的法規越重要
            ls = sorted(ls, key=len)
            return [prefix, *ls]
    return []

if __name__ == '__main__':
    pass
    l = 法條展開('消費者保護法19')
    print(l)
    # 顯示關鍵字查詢法規結果(['政府會計年度']) 



from lark import Lark, Transformer
from pathlib import Path
import logging
logger = logging.getLogger(Path(__file__).stem)
TITLE = -1
PROPOSER = -2
EVENTS = -3

def 列印(docs):
    from collections.abc import Iterable 
    import win32com.client as win32
    from collections.abc import Iterable 
    if isinstance(docs, str) or not isinstance(docs, Iterable):
        docs = [docs]
    word = win32.Dispatch("Word.Application")
    for docfile in docs:
        print(f'列印{docfile }')
        doc = word.Documents.Open(docfile)
        doc.PrintOut()
        doc.Close(False)
    word.Quit()

def move_table_before(table, paragraph):
    tbl, p = table._tbl, paragraph._p
    tbl.addnext(p)

def make_events_table(doc, title, events):
    from docx.shared import Pt, Inches
    from docx.enum.text import WD_ALIGN_PARAGRAPH
    table = doc.add_table(rows=2, cols=2, style="Table Grid")
    title_row = table.rows[0]
    a, b = title_row.cells[:2]
    a.merge(b)
    p = a.paragraphs[0]
    p.paragraph_format.first_line_indent = Pt(0)
    p.paragraph_format.left_indent = Pt(0)
    p.alignment = WD_ALIGN_PARAGRAPH.CENTER
    t = p.add_run(title)
    t.font.bold = True

    heading_row = table.rows[1].cells

    c = heading_row[0]
    c.width = Inches(0.1)
    p = c.paragraphs[0]
    p.alignment = WD_ALIGN_PARAGRAPH.CENTER
    p.paragraph_format.first_line_indent = Pt(0)
    p.paragraph_format.left_indent = Pt(0)
    t = p.add_run('日期')
    t.font.size = Pt(12)
    t.font.bold = True

    c = heading_row[1]
    c.width = Inches(12)
    p = c.paragraphs[0]
    p.alignment = WD_ALIGN_PARAGRAPH.CENTER
    t = p.add_run('摘要')
    t.font.size = Pt(12)
    t.font.bold = True
    p.paragraph_format.first_line_indent = Pt(0)
    p.paragraph_format.left_indent = Pt(0)

    for event in events:
        add_event_row(table, event)
    p = doc.add_paragraph() 
    doc.add_page_break()
    move_table_before(table, p)

def add_event_row(table, event): 
    from docx.enum.text import WD_LINE_SPACING
    from docx.shared import Pt, Inches
    date = 民國日期(取日期(event[0]))
    event = event[1]
    row = table.add_row().cells
    c = row[0]

    p = c.paragraphs[0]
    p.alignment = WD_ALIGN_PARAGRAPH.CENTER
    p.paragraph_format.line_spacing_rule = WD_LINE_SPACING.EXACTLY
    p.paragraph_format.line_spacing = Pt(23)
    t = p.add_run(date)
    t.font.size = Pt(12)

    c = row[1]
    p = c.paragraphs[0]
    p.paragraph_format.line_spacing_rule = WD_LINE_SPACING.EXACTLY
    p.paragraph_format.line_spacing = Pt(23)
    t = p.add_run(event)
    t.font.size = Pt(12)

class MakeDocxTree(Transformer):

    def DESC(self, args):
        return args.value

    def DATE(self, args):
        from zhongwen.date import 取日期
        return 取日期(args)

def setup_style(docx):
    from docx.shared import Pt
    style = docx.styles['Heading 1']
    font = style.font
    font.name = 'Calibri' #'DFKai-SB' #標楷體
    font.size = Pt(16)
    font.bold = True
    #docx.styles['Heading 1'].font.Size = Pt(30)
    return docx

def read_docx(doc):
    '讀取word文件文字'
    from docx import Document
    doc = Document(doc)
    return "\n".join([p.text for p in doc.paragraphs])

def fread_docx(doc):
    from file import 最新符合檔 
    doc = 最新符合檔(Path("D:\\g"), f"**/*{doc}*.docx")
    return read_docx(doc)

def 轉換游標標號層次(層級):
    '游標處符合模式字串，否傳回None'
    import vim
    cb = vim.current.buffer
    l = vim.current.window.cursor[0] - 1
    line = cb[l]
    line = 設定標號層次(層級, line)
    cb[l] = line

def 游標模式替換(模式, 替換函式):
    '游標處符合模式字串，否傳回None'
    import vim
    cb = vim.current.buffer
    _a, l, c, _a, _a = map(lambda s: int(s)-1, vim.eval('getcursorcharpos()'))
    line = cb[l] 
    import re
    p = re.compile(模式)
    i = 0
    while m:=p.search(line, i):
        if m.start() <= c < m.end():
            替換 = 替換函式(m[0]) 
            cb[l] = line[:m.start()] + 替換 + line[m.end():]
            return [m.group(0), m.start(), m.end(), line, cb, l]
        i = m.end()+1
    raise ValueError(f'游標處無此模式，模式:{模式}，游標:{c}，行:{line}')

微軟辦公室軟體共用範本路徑 = [
        'AppData/Roaming/Microsoft/Templates/Normal.dotm' # Word
       ,'AppData/Roaming/Microsoft/Excel/XLSTART/fhopecc.xlsb' # Excel
       ]

def 設定微軟辦公室軟體共用範本():
    from shutil import copy
    logger.info('設定微軟辦公室軟體共用範本……')
    for t in 微軟辦公室軟體共用範本路徑:
        t = Path.home() / t
        s = Path(__file__).parent / 'resource' / t.name
        try:
            copy(s, t)
        except FileNotFoundError:
            t.parent.mkdir(exist_ok=True)
            copy(s, t)
    logger.info('完成。')

def 轉文字檔(filepath):
    from pathlib import Path
    import os
    output_txt = Path(filepath).with_suffix('.txt')
    cmd = f'pandoc -t plain "{filepath}" -o "{output_txt}"'
    os.system(cmd)

def 複製文字(filepath):
    '複製文字至剪貼簿'
    import clipboard
    with open(args.copy_text, 'r', encoding='utf8') as f:
        clipboard.copy(f.read())
    print(clipboard.paste())

def 設定環境():
    from zhongwen.winman import 增加檔案右鍵選單功能, 建立傳送到項目
    from zhongwen.office_document import 設定微軟辦公室軟體共用範本
    from shutil import copy
    import sys
    設定微軟辦公室軟體共用範本()

    cmd =  f'"{sys.executable}" -m zhongwen.office_document --print %* && pause'
    建立傳送到項目('列印微軟文件', cmd)

    cmd = f'{sys.executable} -m zhongwen.office_document --doc2pdf "%1"' 
    增加檔案右鍵選單功能('2pdf', cmd, 'Word.Document.8')  # .doc
    增加檔案右鍵選單功能('2pdf', cmd, 'Word.Document.12') # .docx

    cmd = f'{sys.executable} -m zhongwen.office_document --doc2docx "%1"' 
    增加檔案右鍵選單功能('2docx', cmd, 'Word.Document.8')

    cmd = f'{sys.executable} -m zhongwen.office_document --to_text "%1"' 
    增加檔案右鍵選單功能('2txt', cmd, 'Word.Document.8')
    增加檔案右鍵選單功能('2txt', cmd, 'Word.Document.12') # .docx

    cmd = f'{sys.executable} -m zhongwen.office_document --copy_text "%1"' 
    增加檔案右鍵選單功能('複製文字', cmd, 'Word.Document.8') # .docx
    增加檔案右鍵選單功能('複製文字', cmd, 'Word.Document.12') # .docx

    cmd = f'{sys.executable} -m zhongwen.office_document --save_highlight "%1"' 
    增加檔案右鍵選單功能('另存醒目文字', cmd, 'Word.Document.8') # .docx
    增加檔案右鍵選單功能('另存醒目文字', cmd, 'Word.Document.12') # .docx
    
    cmd = f'{sys.executable} -m zhongwen.office_document --md2docx "%1"' 
    增加檔案右鍵選單功能('markdown2docx', cmd, '.md') # .docx

def 更新微軟辦公室軟體共用範本():
    from shutil import copy
    for s in 微軟辦公室軟體共用範本路徑:
        s = Path.home() / s
        t = Path(__file__).parent / 'resource' / s.name
        copy(s, t)
    logger.info('更新微軟辦公室軟體共用範本完成！')

class MakeContentList(Transformer):
    '回傳[階層碼、段落文字]清單'

    def start(self, args):
        return [[TITLE, args[0].value], *args[1:]]

    def content(self, args):
        return args[0]

    def h1(self, args):   
        from zhongwen.number import 中文數字, 轉數值
        return [1, f"{大寫中文數字(轉數值(args[0].value))}、" + args[1].value]

    def h2(self, args):   
        from zhongwen.number import 中文數字, 轉數值
        return [2, f"({大寫中文數字(轉數值(args[0].value))})" + args[1].value]

    def h3(self, args):   
        from zhongwen.number import 中文數字, 轉數值
        return [3, f"{中文數字(轉數值(args[0].value))}、" + args[1].value]

    def h4(self, args):   
        from zhongwen.number import 中文數字, 轉數值
        return [4, f"({中文數字(轉數值(args[0].value))})" + args[1].value]

    def h5(self, args):   
        from zhongwen.number import 中文數字, 轉數值
        return [5, f"{轉數值(args[0].value)}." + args[1].value]

    def h6(self, args):   
        from zhongwen.number import 中文數字, 轉數值
        return [6, f"({轉數值(args[0].value)})" + args[1].value]

    def p(self, args):   
        return [0, args[0].value]
    
    def proposer(self, args):   
        return [PROPOSER, '提案人：' + args[0].value]
    
    def events(self, args):
        return [EVENTS, [args[0], args[1:]]]

    def event(self, args):
        date = 民國日期(args[0])
        event = args[1]
        return [date, event]

def parse(doc) -> Lark: 
    p = Lark.open(Path(__file__).parent / 'office_document.lark')
    t = p.parse(doc)
    t = MakeDocxTree().transform(t)
    return t

def todocx(doc):
    from docx.shared import Pt, Inches
    from docx.enum.text import WD_ALIGN_PARAGRAPH
    from docx import Document
    path = Path(doc)
    with open(doc, 'r', encoding='utf8') as f:
        doc = '\n'.join(f.readlines())
        docx = path.with_suffix('.docx')
        t = parse(doc)
        title = t.children[0]
        # print(t.pretty())
        t = MakeContentList().transform(t)
        doc = Document(r'd:\GitHub\zhongwen\zhongwen\template.docx')
        for l, p in t:
            if l > 0:
                doc.add_heading(p, level=l)
            else:
                if l == TITLE:
                    p0 = doc.paragraphs[0]
                    p0.alignment = WD_ALIGN_PARAGRAPH.CENTER
                    p0.keep_together = True
                    # p0.keep_with_next = True
                    r = p0.add_run(p)
                    r.font.bold = True
                elif l == PROPOSER: 
                    p = doc.add_paragraph(p) 
                    p.alignment = WD_ALIGN_PARAGRAPH.RIGHT
                elif l == EVENTS:
                    title = p[0]
                    events = p[1]
                    make_events_table(doc, title, events)
                else:
                    doc.add_paragraph(p)
        if '工作底稿' in title:
            doc.add_paragraph()
            doc.add_paragraph("撰稿：                        覆核：")
        doc.save(docx) 
        from os import system
        system(f'start {docx}')

def 合併文件(文件集, 合併文件檔=None, 文件以換頁符號分隔=True):
    '合併微軟辦公室文件(Word)，僅副檔名為 docx 者。' 
    from docxcompose.composer import Composer
    from docx import Document
    from docx.oxml.ns import qn
    from docx.oxml import OxmlElement
    from pathlib import Path
    文件集 = sorted(文件集)

    if not 合併文件檔:
        合併文件檔 = Path(文件集[0]).parent / '合併文件.docx'
    
    input_paths = 文件集
    # Create a new document based on the first input document
    merged_document = Document(input_paths[0])
    composer = Composer(merged_document)

    def 插入換頁符號(document):
        # Create a page break element
        page_break = OxmlElement('w:br')
        page_break.set(qn('w:type'), 'page')
        # Add the page break to the last paragraph
        document.paragraphs[-1]._p.append(page_break)

    for input_path in input_paths[1:]:
        if 文件以換頁符號分隔:
            插入換頁符號(merged_document)
        # Append each subsequent document
        composer.append(Document(input_path))

    # Save the merged document to the specified output path
    composer.save(str(合併文件檔))

def doc2docx(docs, outputdir=None):
    '微軟辦公室文件為 doc 格式者，轉成 docx 格式。'
    from collections.abc import Iterable 
    import comtypes.client
    import pypandoc
    import os
    if isinstance(docs, str) or not isinstance(docs, Iterable):
        docs = [docs]
    for doc in docs:
        input_path = str(doc)
        if outputdir:
            output_path = str(Path(outputdir) / Path(doc).with_suffix('.docx').name)
        else:
            output_path = str(Path(doc).with_suffix('.docx'))
        # Ensure the input file is a .doc file
        if not input_path.lower().endswith('.doc'):
            raise ValueError("Input file must be a .doc file")
        # Ensure the output file is a .docx file
        if not output_path.lower().endswith('.docx'):
            raise ValueError("Output file must be a .docx file")

        # Create a Word application object
        word = comtypes.client.CreateObject('Word.Application')
        word.Visible = False

        # Open the .doc file
        doc = word.Documents.Open(input_path)

        # Save the file as .docx
        doc.SaveAs(output_path, FileFormat=16)  # 16 corresponds to the wdFormatDocumentDefault format

        # Close the document and quit the application
        doc.Close()
        word.Quit()

def doc2pdf(words, output_dir=None):
    from collections.abc import Iterable 
    from pathlib import Path
    import win32com.client
    if isinstance(words, str) or not isinstance(words, Iterable):
        words = [words]
    word = win32com.client.Dispatch('Word.Application')
    for w in words:
        w = Path(w)
        if output_dir:
            pdf = output_dir / w.with_suffix('.pdf').name
        else:
            pdf = w.with_suffix('.pdf')
        doc = word.Documents.Open(str(w))
        doc.SaveAs(str(pdf), FileFormat=17)
        doc.Close()
    word.Quit()

def 標題階層編號轉中文編號(標題):
    from zhongwen.number import 中文數字, 大寫中文數字
    import re
    s = 標題
    pat = r'((\d+.)*(\d+)[\t\s])(.*)'
    if m:=re.match(pat, s):
        階層 = m[1].count('.')+1
        編號 = m[3]
        內容 = m[4]
        if 階層==1:
            編號 = f'{大寫中文數字(編號)}、'
        elif 階層==2: 
            編號 = f'（{大寫中文數字(編號)}）'
        elif 階層==3: 
            編號 = f'{中文數字(編號)}、'
        elif 階層==4: 
            編號 = f'（{中文數字(編號)}）'
        elif 階層==5: 
            編號 = f'{編號}.'
        elif 階層==6: 
            編號 = f'({編號})'
        return 編號 + 內容
    return s

def markdown2docx(md):
    from pathlib import Path
    import win32com.client
    import pypandoc
    import os
    md = Path(md)
    docx = Path(__file__).parent / md.with_suffix('.docx')
    temp = r'd:\GitHub\zhongwen\zhongwen\resource\審核報告範本.docx'
    pypandoc.convert_file(md
                         ,'docx'
                         ,format='markdown+east_asian_line_breaks'
                         ,outputfile=docx
                         ,extra_args=[f"--reference-doc={temp}"
                                     ,"--number-sections"
                                     ]
                         )

    word_app = win32com.client.Dispatch("Word.Application")
    doc = word_app.Documents.Open(str(docx))
    word_app.Visible = True
    try:
        word_app.Run('標題階層編號轉中文編號')
    except Exception as e:
        print(f"執行 VBA 宏時出現錯誤: {e}")
    doc.Save()
    doc.Close()
    word_app.Quit()
    os.system(f'start {docx}')

def 另存醒目文字(docx):
    from docx import Document
    from pathlib import Path
    docx = Path(docx)
    doc = Document(str(docx))

    highlighted_texts = []

    for para in doc.paragraphs:
        本段落前累積高亮區塊數 = len(highlighted_texts)
        for run in para.runs:
            if run.font.highlight_color:  
                highlighted_texts.append(run.text)
        if len(highlighted_texts) > 本段落前累積高亮區塊數:
            highlighted_texts.append('\n')

    highlighted_texts = ''.join(highlighted_texts )
    highlighted_texts = f'宣讀{highlighted_texts }'
    print(highlighted_texts)
    txt = docx.with_suffix('.txt')
    txt.write_text(highlighted_texts, encoding='utf8')

if __name__ == '__main__':
    import argparse
    import clipboard
    logging.basicConfig(level=logging.INFO)
    parser = argparse.ArgumentParser()
    parser.add_argument("--setup", help="設定環境", action="store_true")
    parser.add_argument('--print', nargs='+', type=str, help='列印微軟文件')
    parser.add_argument("--update_temp", help="更新微軟辦公室軟體共用範本", action="store_true")
    parser.add_argument("--doc2pdf", type=str, help="轉 PDF 格式。", required=False)
    parser.add_argument("--doc2docx", type=str, help="doc 轉 docx 格式。", required=False)
    parser.add_argument("--md2docx", type=str, help="markdown 轉 docx 格式。", required=False)
    parser.add_argument("--to_text", type=str, help="轉文字檔", required=False)
    parser.add_argument("--copy_text", type=str, help="複製文字", required=False)
    parser.add_argument("--level_number_to_chinese_number", help="標題階層編號轉中文編號")
    parser.add_argument("--save_highlight", help="另存醒目文字")

    args = parser.parse_args()
    if args.setup:
        設定環境()
    elif docs := args.print:
        列印(docs)
    elif args.update_temp:
        更新微軟辦公室軟體共用範本()
    elif args.doc2pdf:
        doc2pdf(args.doc2pdf)
    elif args.doc2docx:
        doc2docx(args.doc2docx)
    elif args.md2docx:
        markdown2docx(args.md2docx)
    elif args.to_text:
        轉文字檔(args.to_text)
    elif args.copy_text:
        複製文字(args.copy_text)
    elif args.level_number_to_chinese_number:
        print(標題階層編號轉中文編號(args.level_number_to_chinese_number))
    elif args.save_highlight:
        print(另存醒目文字(args.save_highlight))



import unittest

class Test(unittest.TestCase):

    def test_series(self):
        from zhongwen.時序分析 import 最近缺失序列, 序列缺項補零
        import pandas as pd
        s = pd.Series([0.9, 0.9, 1.3, 1.8, 1.4], index=[106, 108, 109, 110, 112])
        年數, 前幾年 = 最近缺失序列(s)
        self.assertEqual((年數, 前幾年), (111, 1))
        完整序列 = 序列缺項補零(s)
        self.assertEqual(完整序列.index.to_list(), [106, 107, 108, 109, 110, 111, 112])
        self.assertEqual(完整序列.to_list(), [0.9, 0.0, 0.9, 1.3, 1.8, 0.0, 1.4])
        s = pd.Series([0.9, 0.9, 1.3, 1.8, 1.4], index=[106, 107, 108, 109, 110])
        年數, 前幾年 = 最近缺失序列(s)
        self.assertEqual((年數, 前幾年), (None, 0))

    def test(self):
        from zhongwen.時序分析 import 最近連續非零正數次數, 連續消長次數
        import pandas as pd
        s = pd.Series([0, 1, 2, 0])
        self.assertEqual(最近連續非零正數次數(s), 0)
        s = [1, 1, 2, 0, 1]
        self.assertEqual(最近連續非零正數次數(s), 1)
        s = [1, 1, 2, 0, 1, 1, 2]
        self.assertEqual(最近連續非零正數次數(s), 3)
        s = pd.Series([1, 1, 2, 0, 1, 1, 2])
        self.assertEqual(最近連續非零正數次數(s), 3)

        s = pd.Series([1, 2, 3, 2, 1, 2, 3])
        self.assertEqual(連續消長次數(s), 2)

    def test_autocorr(self):
        # import pandas as pd
        # import matplotlib.pyplot as plt
        # s = pd.Series([1, 2, 3, 5, 7, 1, 2, 3, 5, 7, 1, 2, 3, 5, 7])
        # print(s.autocorr(2))
        # ax = plt.plot(s.autocorr())
        # s.plot()
        # plt.show()

        import pandas as pd
        import matplotlib.pyplot as plt
        import numpy as np

        # 生成一個時間序列
        # df = pd.DataFrame({
            # "t": range(1, 1001),
            # "y": 0.5 * np.sin(2 * np.pi * t / 24) + 0.5
        # })

        # 計算自相關係數
        # acf = df["y"].autocorr(lag=None)

        # 繪製自相關係數圖像
        # plt.plot(acf)
        # plt.show()

if __name__ == '__main__':
    unittest.main()
    suite = unittest.TestSuite()
    suite.addTest(Test('test_series'))  # 指定測試
    unittest.TextTestRunner().run(suite)



import unittest

class Test(unittest.TestCase):

    def test_markdown(self):
        from zhongwen.office_document import 標題階層編號轉中文編號 
        from subprocess import check_output
        t = '6	擬議處理意見'
        o = '陸、擬議處理意見'
        self.assertEqual(標題階層編號轉中文編號(t), o)

        t = '1.2.1 三級管控核有缺失'
        o = '一、三級管控核有缺失'
        self.assertEqual(標題階層編號轉中文編號(t), o)

        out = check_output('py -m zhongwen.office_document --level_number_to_chinese_number "1.2.1 三級管控核有缺失"', shell=True)
        self.assertEqual(out.decode('cp950').rstrip(), '一、三級管控核有缺失')

    def test(self):
        from zhongwen.office_document import parse
        doc = '''工作底稿
花蓮漁港觀光漁市公共設施新建工程大事記

105.4.21.該規劃於外泊地南側腹地作為上架場用地，惟土地上既有鐵皮屋搭建而成之假日魚市，規劃於106年底拆遷至新設觀光魚市，公告花蓮漁港區域及漁港計畫。

106.5.9.行政院花東地區發展推動小組第14次委員會議決議，花東第二期(105-108 年)綜合發展實施方案滾動檢討案，核列花蓮縣漁港機能改善與多元化利用建置計畫為C類計畫，總經費6,011萬元。

花蓮漁港漁業公共設施新建工程大事記
107.3.27.決標予富太營造股份有限公司。
108.5.13.該府辦理本案驗收結算結果，逾期18天，逾期違約金19萬餘元，結算金額1,071萬餘元。
'''
        t = parse(doc)
        print(t)
        print(t.pretty())
        self.assertEqual(t.children[1].children[0].children[0], '花蓮漁港觀光漁市公共設施新建工程大事記')
        # self.assertEqual(t.children[1].children[1].children[0], '花蓮漁港漁業公共設施新建工程大事記')

        doc = '''肆、調查重點：
一、花蓮縣政府辦理漁港機能改善與多元化利用建置計畫之規劃設計與經費核定及修正情形。

二、預算編列與執行檢核情形。

三、中央補助計畫機關核定之後續規劃及工程執行情形。

四、漁港設施竣工後之管理使用效益。
'''
        # t = parse(doc)
        # print(t.pretty())
        # print(t.children[0])
        # self.assertEqual(t.children[0].children[0].data.value, 'h1')
        # self.assertEqual(t.children[1].children[0].data.value, 'h3')

if __name__ == '__main__':
    # unittest.main()
    suite = unittest.TestSuite()
    suite.addTest(Test('test_markdown'))  # 指定要執行的測試方法
    unittest.TextTestRunner().run(suite)



'檔案處理'
from zhongwen.batch_data import 通知執行時間
from functools import lru_cache
from diskcache import Cache
from pathlib import Path
import logging
cache = Cache(Path.home() / 'cache' / 'zhongwen.file')
logger = logging.getLogger(Path(__file__).stem)

def 轉換(來源:Path, 目的格式):
    '格式以副檔名推論'
    from pydub import AudioSegment
    s = 來源
    來源格式 = s.suffix
    t = s.with_suffix(目的格式)
    if s.suffix == '.aac':
        i = AudioSegment.from_file(s, format=s.suffix.replace('.', ''))
        if 目的格式 == '.mp3':
            o = i.export(t, format=目的格式.replace('.', ''))
            return t 
    raise NotImplementedError(f'尚未實作將{來源}轉換成{目的格式}格式功能')

class FileLocation:
    '萃取文字內之路徑資訊'
    模式集 ={"python":r'File "(?P<path>.+.py)", line (?P<line>\d+).*'
            ,"python_warn":r'^(?P<path>.+.py):(?P<line>\d+):.*'
            ,"python_debugger":r'^> (?P<path>.+.py)\((?P<line>\d+)\)'
            ,"jest":r'\((?P<path>.+\.js):(?P<line>\d+):(?P<pos>\d+)\)'
            ,"path":r'(?P<path>[^"\']+\.(js|py))'
            }
    def __init__(self, 訊息):
        import re 
        # if '股票估值' in 訊息:breakpoint()
        for k in self.模式集:
            模式=self.模式集[k]
            if m:=re.search(模式, 訊息):
                try:
                    self.路徑 = m['path']
                    self.列 = int(m['line'])
                    self.行 = int(m['pos'])
                    break
                except IndexError: pass
                break
        if not hasattr(self, '路徑'): raise ValueError(f'訊息："{訊息}"不包含路徑資訊！')
        if not hasattr(self, '列'): self.列 = 0
        if not hasattr(self, '行'): self.行 = 0
        

def 最新檔(目錄, 檔案樣式="*"):
    import os
    fs = [f for f in 目錄.glob(檔案樣式)]
    try:
        return max(fs, key=os.path.getmtime)
    except:
        raise FileNotFoundError(目錄)

@lru_cache
def chrome():
    # from webdriver_manager.chrome import ChromeDriverManager
    # path = ChromeDriverManager().install()
    from selenium import webdriver
    return webdriver.Chrome()

# @cache.memoize(expire=100, tag='抓取')
def 抓取(url:str
        ,抓取方式='get'
        ,headers=None
        ,回傳資料形態='str'
        ,參數={}
        ,除錯=False
        ,資料={}
        ,encoding="utf-8"
        ,等待秒數=5
        ,return_json=False
        ,return_content=False
        ,return_bytes=False
        ,use_requests=None
        ):
    '''「抓取」網頁內容，傳回字串，惟鍵結以 .xls 或 .xlsx 結尾，視同 Excel 檔，傳回位元組。
另再就抓取網頁內容之鏈結再進行抓取者，稱「爬取」。
抓取方式：'get' 指定使用 requests.get；'post' 係 requests.post；'selenium' 係 selenium 模組。
回傳資料形態: 'str' 傳回字串、'json' 傳回 JSON 物件、'bytes' 傳回位元組。
'''
    from warnings import warn
    from faker import Faker
    import requests
    import logging
    import time

    # 鍵結以 .xls 或 .xlsx 結尾，視同 Excel 檔，傳回位元組。
    if not return_content:
        for suffix in ['.xls', '.xlsx']:
            if url.endswith(suffix):
                return_content=True
                break

    if 抓取方式=='request':
        warn(f'為強化命名，抓取方式參數之【request】選項將廢棄，請以【get】 替代。'
            ,DeprecationWarning, stacklevel=2)
        抓取方式='get'

    if 參數:
        warn(f'為強化命名，【參數】參數項將廢棄，請以【資料】替代。'
            ,DeprecationWarning, stacklevel=2)
        資料=參數

    if return_bytes:
        warn(f'【return_bytes】參數項將廢棄，請將【回傳資料形態】參數設定【bytes】值替代。'
            ,DeprecationWarning, stacklevel=2)
        回傳資料形態 = 'bytes'

    if return_content:
        warn(f'【return_content】參數項將廢棄，請將【回傳資料形態】參數設定【bytes】值替代。'
            ,DeprecationWarning, stacklevel=2)
        回傳資料形態 = 'bytes'

    if return_json:
        warn(f'【return_json】參數項將廢棄，請將【回傳資料形態】參數設定【json】值替代。'
            ,DeprecationWarning, stacklevel=2)
        回傳資料形態 = 'json'

    if use_requests: 
        warn(f'預設使用【requests】模組，【use_request】參數項已無作用並將廢棄。'
            ,DeprecationWarning, stacklevel=2)

    if 抓取方式 == 'selenium':
        c = chrome()
        c.get(url)
        time.sleep(等待秒數)
        if '使用支援JavaScript' in c.page_source:
            c.get(url)
            time.sleep(等待秒數)
        return c.page_source
    if not headers:
        fake = Faker()
        headers = {'user-agent': fake.user_agent()
                  ,"accept-language": "zh-TW,zh;q=0.9,en;q=0.8,zh-CN;q=0.7"
                  }
        if 抓取方式=='post':
            headers["content-type"] = "application/x-www-form-urlencoded"
    if 抓取方式=='post':
        logging.debug(f'「資料」內容：{資料!r}')
        r = requests.post(url, headers=headers, data=資料)
    else:
        r = requests.get(url, headers=headers)
    r.encoding = encoding
    r.raise_for_status()  # 確保請求成功

    除錯訊息 = (f'回復內容為「{r!r}」：\n'
                f'{r.text!r}'
               )
    logging.debug(除錯訊息)

    if 回傳資料形態=='json':
        return r.json()
    elif 回傳資料形態=='bytes':
        return r.content
    else:
        return r.text

def 下載(url, 儲存路徑=None, 儲存目錄=None, 覆寫=False):
    '''下載 URL 內容至指定檔案，並且回傳檔案路徑。'''
    from urllib.parse import urlparse
    import requests
    p = 儲存路徑
    downloads = 儲存目錄
    重載 = 覆寫

    if not downloads:
        downloads = Path.home() / 'Downloads'
        downloads.mkdir(exist_ok=True)
    fn = urlparse(url).path.split('/')[-1]
    if not p:
        p = downloads / fn

    if not 覆寫 and p.exists(): 
        logger.warn(f'警告：[{url}]已下載至[{p}]！')
        return p

    if p.exists():
        p.unlink()
    response = requests.get(url)
    if response.status_code == 200:
        with open(p, 'wb') as file:
            file.write(response.content)
        logger.info(f'下載[{url}]至[{p}]成功！')
    else:
        raise RuntimeError(r"{url}下載失敗，原因如次：", response.status_code, response.reason)
    return p

@通知執行時間
def 解壓(壓縮檔, 目錄):
    if 壓縮檔.suffix == '.7z':
        import py7zr
        壓縮檔 = py7zr.SevenZipFile(壓縮檔, mode='r')
    else:
        import zipfile
        try:
            壓縮檔 = zipfile.ZipFile(壓縮檔)
        except zipfile.BadZipFile:
            raise IOError(f'{壓縮檔}非ZipFile格式')
    壓縮檔.extractall(目錄)
    print(f'解壓[{壓縮檔}]成功！')

def 下載跳出對話視窗連結檔案(url, 目錄=None):
    import requests
    from urllib.parse import urlparse
    import os
    
    response = requests.head(url)
    if 'Content-Disposition' in response.headers:
        content_disposition = response.headers['Content-Disposition']
        filename = content_disposition.split('filename=')[1].strip('"')
    else:
        filename = os.path.basename(urlparse(url).path)
    filepath = Path.home() / 'TEMP' / filename
    response = requests.get(url)
    with open(filepath, 'wb') as file:
        file.write(response.content)
    logger.info(f'文件已下載到：{filepath}')
    return filepath

def 轉文字檔(filepath):
    from pathlib import Path
    import os
    output_txt = Path(filepath).with_suffix('.txt')
    cmd = f'pandoc -t plain "{filepath}" -o "{output_txt}"'
    os.system(cmd)
    return output_txt

def 複製檔案文字至剪貼簿(filepath):
    '複製檔案文字至剪貼簿'
    from pathlib import Path
    import clipboard
    filepath = Path(filepath)
    if filepath.suffix == '.txt':
        with open(filepath, 'r', encoding='utf8') as f:
            clipboard.copy(f.read())
    elif filepath.suffix == '.pdf':
        import pdfplumber
        text = ""
        with pdfplumber.open(filepath) as pdf:
            for page in pdf.pages:
                text += page.extract_text()
        clipboard.copy(text)
    else:
        textfile = 轉文字檔(filepath)
        複製檔案文字至剪貼簿(textfile)

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--copytext", help='複製檔案文字至剪貼簿')
    parser.add_argument("--file2text", help='轉文字檔')
                       
    args = parser.parse_args()
    if file := args.copytext:
        複製檔案文字至剪貼簿(file)
    elif file := args.file2text:
        轉文字檔(file)



'windows 系統管理工具'
from pathlib import Path
import os
import logging
logger = logging.getLogger(Path(__file__).stem)

def setx(var, value):
    '設定 Windows 環境變數'
    #cmd = f'setx {var} {value} /M >nul 2>nul'
    cmd = f'setx {var} "{value}"'
    if not (r:=os.system(cmd)) == 0:
        raise WindowsError(f'設定 Windows 環境變數指令[{cmd}]失敗，回傳值為[{r}]！')
    print(f'設定 Windows 環境變數[{var}]為[{value}]成功！')

def addpath(p, path_var='PATH'):
    p = str(p)
    ps = os.environ[path_var].split(';')
    if p not in ps:
        p = p.replace('"', '')
        ps.append(p)
        ps = list(set(ps))
        setx(path_var, ';'.join(ps))

TEMP = Path(os.environ['TEMP'])

def downloads():
    return Path.home() / 'Downloads'

def services():
    resume = 0
    accessSCM = win32con.GENERIC_READ
    accessSrv = win32service.SC_MANAGER_ALL_ACCESS

    #Open Service Control Manager
    hscm = win32service.OpenSCManager(None, None, accessSCM)

    #Enumerate Service Control Manager DB
    typeFilter = win32service.SERVICE_WIN32
    stateFilter = win32service.SERVICE_STATE_ALL

    statuses = win32service.EnumServicesStatus(hscm, typeFilter, stateFilter)

    return [short_name for (short_name, desc, status) in statuses]

def 是否安裝(software):
    from windows_tools.installed_software import get_installed_software
    for s in [s['name'] for s in get_installed_software()]:
        if software in s:
            return True
    return False

def powershell(cmd):
    import subprocess
    result = subprocess.run(["powershell", "-Command", cmd], capture_output=True)
    if result.returncode !=0:
        raise WindowsError(f'執行 powershell 發生錯誤：{result}；指令{cmd}')

def remove_cortan():
    cmd = 'Get-AppxPackage -allusers Microsoft.549981C3F5F10 | Remove-AppxPackage'        
    powershell(cmd)    

def 建立傳送到項目(名稱:str, 命令:str):
    import os
    import winshell
    批次檔 = Path(winshell.sendto()) / f"{名稱}.bat"
    批次檔.write_text(命令)
    logger.info(f'建立傳送到【{名稱}】。')

def 增加檔案右鍵選單功能(功能名稱:str, 指令:str, 副檔名='*'):
    import winreg as reg
    import os
    import sys

    # 功能名稱
    context_menu_name = 功能名稱
    
    key_path = fr'{副檔名}\shell\{context_menu_name}'
    command_key_path = fr'{副檔名}\shell\{context_menu_name}\command'

    # 創建註冊表項目
    try:
        reg.CreateKey(reg.HKEY_CLASSES_ROOT, key_path)
        reg.CreateKey(reg.HKEY_CLASSES_ROOT, command_key_path)
        
        # 設置右鍵選單項目名稱
        with reg.OpenKey(reg.HKEY_CLASSES_ROOT, key_path, 0, reg.KEY_WRITE) as key:
            reg.SetValue(key, '', reg.REG_SZ, context_menu_name)

        # 設置指令
        with reg.OpenKey(reg.HKEY_CLASSES_ROOT, command_key_path, 0, reg.KEY_WRITE) as key:
            reg.SetValue(key, '', reg.REG_SZ, 指令)
        if 副檔名 == '*':
            logger.info(f"所有檔案右鍵選單增加【{功能名稱}】項目。")
        else:
            logger.info(f"{副檔名} 右鍵選單增加【{功能名稱}】項目。")
    except Exception as e:
        logger.info(f"添加右鍵選單項目時出錯: {e}")

def where(command):
    '查找指定的可執行檔，以Windows where.exe 實作。'
    import subprocess
    result = subprocess.run(['where', command], capture_output=True, text=True)
    paths = result.stdout.strip().split('\n')
    return paths

def 設定預設右鍵選單():
    import sys
    cmd = f'"{sys.executable}" -m zhongwen.file --copytext "%1"' 
    增加檔案右鍵選單功能('複製檔案文字', cmd, '*')
    cmd = f'"{sys.executable}" -m zhongwen.file --file2text "%1"' 
    增加檔案右鍵選單功能('轉文字檔', cmd, '*')
 
if __name__ == '__main__':
    設定預設右鍵選單()



from pathlib import Path
import logging
import atexit
logger = logging.getLogger(Path(__file__).stem)

def 通知執行時間(f):
    from functools import wraps
    from time import time
    @wraps(f)
    def wrap(*args, **kw):
        ts = time()
        result = f(*args, **kw)
        te = time()
        logger.info(f'{f.__name__}費時{time()-ts:.2f}秒。')
        return result
    return wrap

class 查無批號錯誤(Exception): pass

def 批次讀取(批號, 批號欄名, 表格, 資料庫, 日期欄位=None):
    import pandas as pd
    from zhongwen.date import 全是日期嗎
    # sql 相等運算子係單等號，而 Python 係雙等號，易彼此誤植

    if isinstance(批號, pd.Period):
        批號 = str(批號)
    elif 全是日期嗎(批號):
        批號 = 轉日期字串(批號)

    if isinstance(批號, str):
        sql = f'select * from {表格} where {批號欄名}="{批號}"' 
    else:
        sql = f'select * from {表格} where {批號欄名}={批號}' 

    df = pd.read_sql_query(sql, 資料庫, parse_dates=日期欄位)
    if df.empty: 
        raise 查無批號錯誤(f'查無【{批號}】批號！')
    return df

def 批次刪除(批號, 批號欄名, 表格, 資料庫):
    from zhongwen.date import 全是日期嗎
    import pandas as pd
    logger.debug(f'{表格}刪除{批號}整批紀錄……')
    c = 資料庫.cursor()
    if isinstance(批號, pd.Period):
        批號 = str(批號)
    elif 全是日期嗎(批號):
        批號 = 轉日期字串(批號)
    sql = f'delete from {表格} where {批號欄名}=="{批號}"'
    c.execute(sql)
    資料庫.commit()
    logger.debug(f'成功')

def 批次寫入(資料, 批號, 批號欄名, 表格, 資料庫, 指定欄位=None):
    '如有指定欄位，原始資料則另寫入原始資料庫'
    from zhongwen.text import 臚列
    from zhongwen.date import 有日期嗎
    import pandas as pd
    import sqlite3
    import re

    if 資料.empty:
        return 
    資料.loc[:, 批號欄名] = 批號
    表格不存在=False
    try:
        批次刪除(批號, 批號欄名, 表格, 資料庫)
    except sqlite3.OperationalError as e:
        if 'no such table' in str(e):
            logger.debug(f'尚無【{表格}】資料表……')
            表格不存在=True
        else:
            logger.debug(f'批次刪除發生錯誤：{e}')
    logger.debug(f'批號【{批號}】資料擬整批寫入至{表格}……')
    if 指定欄位:
        if 批號欄名 not in 指定欄位:
            指定欄位.append(批號欄名)
        指定欄位 = [c for c in 資料.columns if c in 指定欄位]
        欄位縮減資料庫 = 資料庫
        資料庫檔路徑 = 取資料庫檔路徑(資料庫)
        資料庫檔路徑 = 資料庫檔路徑.with_stem(f'原始{資料庫檔路徑.stem}')
        資料庫 = 取資料庫(資料庫檔路徑)
        try:
            批次刪除(批號, 批號欄名, 表格, 資料庫)
        except sqlite3.OperationalError as e:
            if 'no such table' in str(e):
                logger.debug(f'原始資料庫尚無【{表格}】資料表……')
                表格不存在=True
            else:
                logger.debug(f'批次刪除發生錯誤：{e}')
        批次寫入(資料[指定欄位], 批號, 批號欄名, 表格, 欄位縮減資料庫)
    try:
        for c in 資料.columns:
            if 有日期嗎(資料[c]):
                資料[c] = 資料[c].map(轉日期字串)
            elif isinstance(資料[c].dtype, pd.PeriodDtype):
                資料[c] = 資料[c].map(str)
        資料.to_sql(表格, 資料庫, if_exists='append')
        if 表格不存在:
            logger.debug(f'擬建立表格【{表格}】及其索引【{批號欄名}】……')
            sql = f'CREATE INDEX idx_{批號欄名} ON {表格}({批號欄名})'
            cursor = 資料庫.cursor()
            cursor.execute(sql)
            資料庫.commit()
            logger.debug('完成！')
        logger.debug(f'寫入成功！')
    except sqlite3.OperationalError as e:
        df = pd.read_sql_query(f'select * from {表格} limit 1', 資料庫)
        寫入資料欄位 = 資料.columns.to_list()
        資料庫欄位 = df.columns.to_list()
        差異欄位 = set(寫入資料欄位) - set(資料庫欄位)
        logger.debug(f'寫入資料之{臚列(差異欄位)}等欄位，資料庫尚未建立……')
        cursor = 資料庫.cursor()
        for c in 差異欄位:
            alter_query = f'ALTER TABLE {表格} ADD COLUMN "{c}" INT'
            cursor.execute(alter_query)
        資料庫.commit()
        logger.debug('前開資料庫未建立欄位已新增……')
        資料.to_sql(表格, 資料庫, if_exists='append')
        logger.debug(f'寫入成功！')

def 期日資料批次寫入(資料庫檔, 資料名稱, 資料日期欄名, 預設資料日期組=[]):
    '''逐批寫入期日資料，資料日期一定為日期，即有year及month方法，舉如：
爬取損益表(資料日期組=迄每季(季末(2019, 1)))
會爬取自2019第1季至今每一季之損益表。
'''
    from collections.abc import Iterable
    from functools import wraps
    from sqlite3 import connect
    def 增加結果批次寫入(爬取期日資料函數):
        @wraps(爬取期日資料函數)
        def 批次寫入爬取資料(資料日期組=None, **kargs):
            if not 資料日期組:
                資料日期組 = 預設資料日期組
            if not isinstance(資料日期組, Iterable):
                資料日期組 = [資料日期組]
            with connect(資料庫檔) as db: 
                for 資料日期 in 資料日期組:
                    try:
                        資料日期.year
                        資料日期.month
                    except AttributeError:
                        raise TypeError(f'資料日期不為日期型態，而為{type(資料日期)}')
                    df = 爬取期日資料函數(資料日期)
                    批次寫入(df, 資料日期, 資料日期欄名, 資料名稱, db)
                return df
        return 批次寫入爬取資料
    return 增加結果批次寫入
    

def 結果批次寫入(資料庫檔, 資料名稱, 批號欄名, 預設批號組=[], 指定欄位=None):
    '''「預設批號組」指定一組批號逐批寫入查詢結果，舉如：
爬取損益表(批號組=迄每季(季末(2019, 1)))
會爬取自2019第1季至今每一季之損益表。
'''
    from collections.abc import Iterable
    from functools import wraps
    def 增加結果批次寫入(爬取資料函數):
        @wraps(爬取資料函數)
        def 批次寫入爬取資料(批號組=None, **kargs):
            if not 批號組:
                批號組 = 預設批號組
            if isinstance(批號組, str) or not isinstance(批號組, Iterable):
                批號組 = [批號組]
            db = 取資料庫(資料庫檔) 
            for 批號 in 批號組:
                df = 爬取資料函數(批號, **kargs)
                if df.empty:
                    continue
                批號 = df.iloc[0][批號欄名]
                批次寫入(df, 批號, 批號欄名, 資料名稱, db, 指定欄位)
            return df
        return 批次寫入爬取資料
    return 增加結果批次寫入

class 批號存在錯誤(Exception):
    def __init__(self, 批號, 批號欄名, 表格):
        self.批號 = 批號
        self.批號欄名 = 批號欄名
        self.表格 = 表格

    def __str__(self):
        return f'批號【{self.批號}】已存在，請指定覆寫=True！'

def 載入批次資料(資料庫檔, 表格, 批次欄名, 時間欄位=None, 期間欄位=None):
    '批次欄名'
    import pandas as pd
    import sqlite3
    from zhongwen.時 import 取日期, 取期間
    from collections.abc import Iterable 
    cs, ps = 時間欄位, 期間欄位
    if isinstance(cs, str) or not isinstance(cs, Iterable):
        cs = [cs]
    if isinstance(ps, str) or not isinstance(ps, Iterable):
        ps = [ps]
    
    with sqlite3.connect(資料庫檔) as c:
        # sql = f"select distinct * from {表格}" # select distinct 將降低效能
        sql = f"select * from {表格}"
        try:
            df = pd.read_sql_query(sql, c, index_col='index') 
        except KeyError as e:
            logger.info(f'{表格}無 index 欄位！')
            df = pd.read_sql_query(sql, c) 
        for c in cs:
            df[c] = df[c].map(取日期)
        for p in ps:
            df[p] = df[p].map(取期間) 
        return df

def 應更新資料時期(更新頻率='次月十日前'):
    from zhongwen.date import 今日, 上月
    if 更新頻率=='次月十日前':
        if 今日().day <= 10:
            return 上月()

def 解析更新期限(更新期限='次月10日前'):
    '解析更新期限之有效值有【每次更新】、【次月10日前】、【次季45日前】、【次月底前】及【財報】，傳回【應更新資料時期、應更新資料期限及定期更新資訊】'
    from zhongwen.date import  上年底, 上月, 上季, 今日, 月底, 取日期, 季別
    from zhongwen.date import  民國正式日期, 民國年月, 民國年, 民國季別
    from datetime import timedelta
    import re
    today = 今日()
    應更新資料時期, 應更新資料期限 = None, None
    if 更新期限=='每次更新':
        return 解析更新期限('次月底前')

    if '財報' in 更新期限:
        '''證交所規定上市公司公告申報財務報表之期限如下：
    一、年度財務報告：每會計年度終了後3個月內(3/31前)。
    二、第一季、第二季、第三季財務報告：
    (一)一般公司(含投控公司)：每會計年度第1季、第2季及第3季終了後45日內(5/15、8/14、11/14前)。
    (二)保險公司：每會計年度第1季、第3季終了後1個月內(4/30、10/31前)，每會計年度第2季終了後2個月內(8/31前)。
    (三)金控、證券、銀行及票劵公司：每會計年度第1季、第3季終了後45日內(5/15、11/14前)，每會計年度第2季終了後2個月內(8/31前)。惟金控公司編製第1季、第3季財務報告時，若作業時間確有不及，應於每季終了後60日內(5/30、11/29前)補正。'''
        應更新資料時期 = 上季()
        _, 季數 = 季別(應更新資料時期)
        match 季數: 
            case 2:
                應更新資料期限 = 取日期(f'{應更新資料時期.year}.8.31')
            case 4:
                應更新資料期限 = 取日期(f'{應更新資料時期.year+1}.3.31')
            case _:
                應更新資料期限 = 應更新資料時期 + timedelta(days=60)
        return (應更新資料時期, 應更新資料期限, f'{民國正式日期()}應公布{民國季別(應更新資料時期)}資訊。')
        
    if '次月' in 更新期限:
        應更新資料時期 = 上月()
    elif '次年' in 更新期限:
        應更新資料時期 = 上年底()
    elif '次季' in 更新期限:
        應更新資料時期 = 上季()

    pat = r'(\d+)(日|個月)(內|前)'
    if m:=re.search(pat, 更新期限):
        n = int(m[1])
        if m[2] == "日":
            應更新資料期限 = 應更新資料時期 + timedelta(days=n)
        elif m[2] == "個月":
            from datetime import datetime, timedelta
            import calendar
            data_date = 應更新資料時期 + timedelta(days=1)
            month = data_date.month-1
            year = data_date.year
            # breakpoint()
            # 計算2個月後的日期
            future_month = month + n
            if future_month > 12:
                future_month -= 12
                year += 1

            # 確定未來月份的天數
            _, days_in_future_month = calendar.monthrange(year, future_month)

            # 設定未來日期
            future_date = data_date.replace(year=year, month=future_month, day=max(data_date.day, days_in_future_month))

            應更新資料期限 = future_date
        if today.month == 2: # 因應2月初通常為春節假期，是以延長更新期限至2月底
            應更新資料期限 = 月底()

    pat = r'月底前'
    if m:=re.search(pat, 更新期限):
        應更新資料期限 = 月底()
    if '次月' in 更新期限:
        return (應更新資料時期, 應更新資料期限, f'{民國正式日期()}應公布{民國年月(應更新資料時期)}資訊。')
    elif '次年' in 更新期限:
        return (應更新資料時期, 應更新資料期限, f'{民國正式日期()}應公布{民國年(應更新資料時期)}資訊。')
    elif '次季' in 更新期限:
        return (應更新資料時期, 應更新資料期限, f'{民國正式日期()}應公布{民國季別(應更新資料時期)}資訊。')

def 增加定期更新(更新期限='次月10日前', 更新程序=解析更新期限):
    '''參數「更新期限」如指定值為「次季45日前」，係指自季初連續45日，另因應公司遇假日未及更新資料加計緩衝期3日，合計48日，執行更新上季資料程序，逾限停更，更多期限如「次月10日前」、「次月底前」……。
查詢函數之參數「更新」如指定為True，則強制更新該批資料。
'''
    from zhongwen.batch_data import 解析更新期限
    from zhongwen.date import 今日
    from datetime import timedelta
    from functools import wraps
    from pathlib import Path
    緩衝期 = timedelta(days=3)
    def 增加定期更新功能(查詢資料):
        @wraps(查詢資料)
        def 查詢定期更新資料(*args, 更新=False,**kargs):
            資料時期, 公布期限, 資訊 = 解析更新期限(更新期限)
            if 更新 or (資料時期 < 今日() <= 公布期限 + 緩衝期):
                更新程序(資料時期)
            return 查詢資料()
        return 查詢定期更新資料
    return 增加定期更新功能

def 去除重覆(資料庫, 表格, 批號, 時間欄位=None):
    import sqlite3
    df = 載入批次資料(資料庫, 表格, 批號, 時間欄位)
    df = df.drop_duplicates()
    with sqlite3.connect(資料庫) as c:
        sql = f'drop table {表格}'
        c.execute(sql)
        c.commit()
        df.to_sql(表格, c)
    logger.info(f'{表格}去除重覆紀錄完成！') 

def 取資料庫檔路徑(conn) -> Path:
    'PRAGMA database_list 查詢資料庫及其儲存檔案資料表，其欄位名稱為 seq, name 及 file 分別表示資料庫序號、名稱及檔案路徑。'
    cursor = conn.cursor()
    cursor.execute('PRAGMA database_list;')
    db_list = cursor.fetchall()
    cursor.close()
    return Path(db_list[0][2])

@通知執行時間
def 取資料表內容(資料庫路徑, 資料表):
    from sqlite3 import connect
    from zhongwen.batch_data import 取資料庫
    import pandas as pd
    sql = f'select * from "{資料表}"'
    return pd.read_sql_query(sql, 取資料庫(資料庫路徑), index_col='index') 

@通知執行時間
def 取原始資料表內容(資料庫路徑, 資料表):
    return 取資料表內容(資料庫路徑.with_stem(f'原始{資料庫路徑.stem}'), 資料表)

def 自原始資料庫重建(資料庫路徑:Path, 資料表, 批號及指定欄位):
    df = 取原始資料表內容(資料庫路徑, 資料表)
    df[批號及指定欄位].to_sql(資料表, 取資料庫(資料庫路徑), if_exists='replace')

def 轉日期字串(日期):
    '日期轉換為格式如"1979-7-29"之字串，而無效日期為"無效日期"字串，俾將日期以字串儲存於 SQLite 資料庫。'
    from zhongwen.date import 全是日期嗎
    if 全是日期嗎(日期):
        return f"{日期.year}-{日期.month}-{日期.day}"
    return "無效日期"
資料庫池 = {}
def 取資料庫(資料庫檔路徑):
    import sqlite3
    try:
        db = 資料庫池[資料庫檔路徑]
        db.execute("select 1")
        return db
    except (KeyError, sqlite3.ProgrammingError):
        db = sqlite3.connect(資料庫檔路徑)
        資料庫池[資料庫檔路徑] = db 
        return db

def 關閉資料庫池():
    for conn in 資料庫池.values(): 
        conn.close()
atexit.register(關閉資料庫池)



import unittest
from pathlib import Path
import pandas as pd
import os
from zhongwen.pandas_tools import show_html

class Test(unittest.TestCase):
    # @unittest.skip('不測試')
    def test(self):
        from zhongwen.花蓮縣法規 import 法規
        # df = cache.get('花蓮縣法規')
        # print(df)
        # for k in cache.iterkeys():
            # print(k)
        # print(cache.iterkeys())
        df = 法規()
        show_html(df)
        # df = df[['法規名稱', '公發布日']]
        # df.reset_index(inplace=True)
        # df = df.iloc[:100]

    def test_parser(self):
        from zhongwen.花蓮縣法規 import 法規分條
        c = "第一條花蓮縣政府(以下簡稱本府)為如期推動污水下水道工程，配合水資源回收中心運轉操作，經公告本縣境內污水下水道聯接使用（以下簡稱用戶接管)工程範圍內違章建築認定與拆遷作業，悉依本準則規定辦理。用戶接管工程範圍外之違章建築，依建築法、違章建築處理辦法及花蓮縣違章建築查報及拆除標準作業程序等相關規定辦理。第二條經公告用戶接管工程範圍內違章建築之拆遷作業，不適用有關補償及救濟法規之規定。第三條施工範圍內違章建築之拆除及施工作業，按下列規定辦理：一、施工單位應於施工前三十日內完成說明會之舉行與反映意見之答覆，並應於開工前十五日會同村里長逐戶張貼通告單，且與有爭議之用戶協調。二、施工範圍由用戶自行興闢留設單側七十五公分空間（雙側一點五公尺）以供作業者，依其留設空間施工，自行興闢留設用戶施工範圍之修復工程，由用戶自行負責。三、惟施工前未留設，經催告十五日後仍未留設者，依本準則查報拆除地面層防火間隔一百五十公分有危害防火逃生部分，及二樓以上有危害建物結構安全之違建，拆除後之修復工程，由用戶自行負責，其程序如附表所載。第四條工程竣工後如施工範圍再行增加違章建築，依第一條第二項規定辦理。第五條本準則自發布日實施。"
        l = 法規分條(c)
        self.assertEqual(l[0][0], 1)
        self.assertEqual(l[4][0], 5)
        self.assertEqual(l[4][1], '本準則自發布日實施。')

        c= "一、花蓮縣政府（以下簡稱本府）為統一審查農舍及農業設施申請建築執照，申請基地未臨接建築線，以其他土地通路或農路臨接道路使用，依內政部77年8月18日台內營字第622851號、內政部營建署99年7月30日營署建管字第0992914816號函意旨，訂定本原則。二、農業用地興建未達建築法第7條規定雜項工作物以外之簡易鋪面（混凝土、瀝青混凝土等類似材料）或透水性鋪面通路及農路，免依建築法規定申請建築執照。三、農舍或農業設施申請建築執照，於基地內興建通路或農路者應檢附農業用地作農業設施容許使用同意文件或經農業主管機關（或公共工程主管機關）認定為公共建設及通行需要所核發證明文件。四、農舍或農業設施申請建築執照以其他土地通路或農路臨接道路使用，除公有土地現況作為道路使用者之外，應取得該土地所有權人通行使用同意書。本府主管建築機關於核發農舍或農業設施建造執照時，除下列情形外應副知本府農業主管機關（都市計畫內土地另副知都市計畫主管機關）列管查處：（一）該通路或農路取得農業用地作農業設施容許使用同意文件。（二）該通路或農路經農業主管機關（或公共工程主管機關）認定為公共建設及通行需要所核發證明文件，且現況可供通行者。（三）該通路或農路範圍編定為交通用地。（四）該通路或農路之使用分區或土地編定為建築用地，符合建築技術規則建築設計施工編第一百六十三條基地內通路規定者。五、依本原則申請建築許可，其臨接通路或農路寬度及退讓應符合其他法令規定，倘涉及他人權益，應依私權及司法途徑調處，建築物起造人、或設計人、或監造人、或承造人，如有侵害他人財產，或肇致危險或傷害他人時，應分別依法負其責任。六、本原則奉縣長核定後公告實施。"
        l = 法規分條(c)
        self.assertEqual(l[0][0], 1)
        self.assertEqual(l[4][0], 5)
        self.assertEqual(l[5][1], '本原則奉縣長核定後公告實施。')
        c = "詳如附表所示。"
        l = 法規分條(c)
        self.assertEqual(l[0][0], 1)
        self.assertEqual(l[0][1], '詳如附表所示。')

if __name__ == '__main__':
    unittest.main()



import unittest
from pathlib import Path

class Test(unittest.TestCase):
    # @unittest.skip('skip')
    def test_company_name(self):
        from zhongwen.text import 公司正名
        n0 = '中國信託商業銀行(股)公司(下稱中信銀行)'
        n1 = '中國信託商業銀行股份有限公司'
        self.assertEqual(公司正名(n0), n1)

    # @unittest.skip('skip')
    def test_moedict(self):
        from zhongwen.text import 查萌典, 萌典尚無定義之字詞
        
        # import logging
        # logging.getLogger().setLevel(logging.DEBUG)

        self.assertEqual(查萌典('查'), ['ㄔㄚˊ：考察、檢查。翻閱、檢尋。大筏，水中的浮木。', 
                                        'ㄓㄚ：姓。如五代時南唐有查文徽。我。同「咱」(一)。'])

        with self.assertRaises(萌典尚無定義之字詞):
            查萌典('word')
        import logging
        logging.getLogger().setLevel(logging.DEBUG)
        self.assertEqual(查萌典('彊')[0], '「強」的異體字。')

    # @unittest.skip('隔離測試')
    def test_text(self):
        from zhongwen.text import 轉全型, 轉半型
        self.assertEqual(轉全型('%'), '％')
        self.assertEqual(轉半型('％'), '%')

        from zhongwen.text import 是否為字元
        self.assertTrue(是否為字元('簡'))
        self.assertTrue(是否為字元('a'))
        self.assertTrue(是否為字元('μ'))
        self.assertFalse(是否為字元('.'))
        self.assertFalse(是否為字元('，'))

        from zhongwen.text import 是否為中文字元, 字元切換
        self.assertTrue(是否為中文字元('簡'))
        self.assertTrue(是否為中文字元('简'))
        self.assertFalse(是否為中文字元('a'))
        self.assertFalse(是否為中文字元('μ'))
        self.assertEqual(字元切換('a'), 'A')
        self.assertEqual(字元切換('A'), 'a')
        self.assertEqual(字元切換('B'), 'b')
        self.assertEqual(字元切換('b'), 'B')
        self.assertEqual(字元切換('簡'), '简')
        self.assertEqual(字元切換('简'), '簡')
        self.assertEqual(字元切換('か'), 'カ')
        self.assertEqual(字元切換('カ'), 'か')
        self.assertEqual(字元切換('['), '「')
        self.assertEqual(字元切換(']'), '」')
        self.assertEqual(字元切換('行'), '行')

        from zhongwen.text import 臚列
        self.assertEqual(臚列(['甲方', '乙方', '丙方']), '甲方、乙方及丙方')
        self.assertEqual(臚列(['單方']), '單方')
        self.assertEqual(臚列('片面'), '片面')

        from zhongwen.text import 倉頡對照表, 倉頡首碼, 首碼搜尋表示式
        m = 倉頡對照表()
        self.assertEqual(m['稜'], 'hdgce')
        # self.assertEqual(m['函'], 'hdgce')
        self.assertEqual(倉頡首碼('稜'), 'h')
        self.assertEqual(倉頡首碼('a'), 'a')
        # self.assertEqual(倉頡首碼('函'), 'u')
        text = '''fa 命令原係向前搜尋字母a，
擴充為向前搜尋字母a及中文倉頡碼首碼為a的中文字，
如【是】倉頡碼為【amyo】。函u
'''
        self.assertTrue('是' in 首碼搜尋表示式('a', text))
        self.assertTrue('倉' in 首碼搜尋表示式('o', text))
        self.assertTrue('令' in 首碼搜尋表示式('o', text))
        self.assertTrue('係' in 首碼搜尋表示式('o', text))

        from zhongwen.text import 翻譯
        self.assertEqual(翻譯('test'), '測試')
        # self.assertEqual(翻譯('取り'), '拿')

    def test_text_segmentation(self):
        from zhongwen.text import 分詞

        字串 = '''美國十年期公債殖利率即你現在買進且持有至到期日每年的年化報酬率，
其係依據債券現價、持有至到期日期間之利息及到期還本現金流所計算之內部報酬率，
由於美國係目前世界上經濟最發達的國家，
其所發行公債倒債風險極低，該公債被大部分投資人視為無風險投資，
是以，美國十年期公債殖利率通常視作無風險利率。
'''
        self.assertTrue('債券' in 分詞(字串))

        # self.assertEqual(中文詞界(10, line), (9,10))

    def test_to_css_string(self):
        from zhongwen.text import 轉樣式表字串
        s = '[長空]：上年度股利估計相對誤差達20.40%；[短多]：毛利率達54.27%，營利率達23.48%'
        r = r'[\009577\007a7a]\00ff1a\004e0a\005e74\005ea6\0080a1\005229\004f30\008a08\0076f8\005c0d\008aa4\005dee\00905420.40%\00ff1b[\0077ed\00591a]\00ff1a\006bdb\005229\007387\00905454.27%\00ff0c\0071df\005229\007387\00905423.48%'
        self.assertEqual(轉樣式表字串(s), r)

    @unittest.skip('尚未完成')
    def test_surround(self):
        from zhongwen.text import 取最近環繞符號, 刪除環繞符號, 插入環繞符號
        from zhongwen.text import 取引用內容, 引用
        s = 'There is a (parentheses) and the nearest (parenthese).'
        self.assertEqual(取引用內容(s), None)
        self.assertEqual(取引用內容(s, 10), None)
        self.assertEqual(取引用內容(s, 11), 引用('(parentheses)', 11, 23))
        self.assertEqual(取引用內容(s, 12), 引用('(parentheses)', 11, 23))
        self.assertEqual(取引用內容(s, 22), 引用('(parentheses)', 11, 23))
        self.assertEqual(取引用內容(s, 23), 引用('(parentheses)', 11, 23))
        self.assertEqual(取引用內容(s, 24), None)
        self.assertEqual(取引用內容(s, 40), None)
        self.assertEqual(取引用內容(s, 41), 引用('(parenthese)', 41, 52))
        self.assertEqual(取引用內容(s, 42), 引用('(parenthese)', 41, 52))
        self.assertEqual(取引用內容(s, 51), 引用('(parenthese)', 41, 52))
        self.assertEqual(取引用內容(s, 52), 引用('(parenthese)', 41, 52))

        s = 'delete (the [neareast surround]).'
        self.assertEqual(取引用內容(s, 31), 引用('(the [neareast surround])', 7, 31))
        self.assertEqual(取引用內容(s, 7), 引用('(the [neareast surround])', 7, 31))
        self.assertEqual(取引用內容(s, 8), 引用('(the [neareast surround])', 7, 31))
        self.assertEqual(取引用內容(s, 11), 引用('(the [neareast surround])', 7, 31))
        self.assertEqual(取引用內容(s, 12), 引用('[neareast surround]', 12, 30))
        self.assertEqual(取引用內容(s, 30), 引用('[neareast surround]', 12, 30))

        s = '這列一個「引號」，再列一個『雙引號』。'
        self.assertEqual(取引用內容(s, 4), 引用('「引號」', 4, 7))
        self.assertEqual(取引用內容(s, 12), None)
        self.assertEqual(取引用內容(s, 13), 引用('『雙引號』', 13, 17))
        self.assertEqual(取引用內容(s, 15), 引用('『雙引號』', 13, 17))
        self.assertEqual(取引用內容(s, 17), 引用('『雙引號』', 13, 17))
        self.assertEqual(取引用內容(s, 18), None)

        s = "刪除「游標處「巢狀引號」的功能」。"
        # self.assertEqual(取引用內容(s, 5), 引用('「游標處「巢狀引號」的功能」', 6, 11))
        # self.assertEqual(取引用內容(s, 6), 引用('「巢狀引號」', 6, 11))
        # self.assertEqual(取引用內容(s, 7), 引用('「巢狀引號」', 6, 11))
        # self.assertEqual(取引用內容(s, 11), 引用('「巢狀引號」', 6, 11))

        # r = "刪除「游標處巢狀引號的功能」。"
        # self.assertEqual(刪除環繞符號(s, "「", 7), r)
        # self.assertEqual(刪除環繞符號(s, "「", 8), r)
        # self.assertEqual(刪除環繞符號(s, "「", 12), r)
        s = "刪除'游標處單引號'的功能"
        r = "刪除游標處單引號的功能"
        # self.assertEqual(刪除環繞符號(s, "'", 3), r)
        # self.assertEqual(刪除環繞符號(s, "'", 4), r)
        # self.assertEqual(刪除環繞符號(s, "'", 9), r)
        # self.assertEqual(刪除環繞符號(s, "'", 10), r)
        s = "'日期欄位'"
        # self.assertEqual(取最近環繞符號(s), "''")
        s = "'日期欄位'"
        # self.assertEqual(刪除環繞符號(s, "'"), "日期欄位")
        t = 'enclose this (cursorword) in parentheses.'
        # self.assertEqual(插入環繞符號('enclose this cursorword in parentheses.', 列=14), t)
      # 測試
    @unittest.skip('尚未完成')
    def test_find_matching_parentheses(self):
        from zhongwen.text import 查找引號對
        expression = "a * (b + c) - {d / [e + f]}"
        matches = 查找引號對(expression)
        print(matches)

if __name__ == '__main__':
    unittest.main()
    suite = unittest.TestSuite()
    suite.addTest(Test('test_find_matching_parentheses'))  # 指定要執行的測試方法
    unittest.TextTestRunner().run(suite)



from unittest.mock import patch
from zhongwen.date import 今日
import unittest

class Test(unittest.TestCase):
    def test_parse_date(self):
        from zhongwen.date import 全是日期嗎, 取日期
        from pandas import Timestamp, Timedelta
        from datetime import datetime, date
        import pandas as pd
        self.assertEqual(取日期("'24/01/19"), Timestamp(2024,1,19))
        self.assertEqual(取日期('2022/6/3 上午 12:00:00'), Timestamp(2022,6,3) )
        self.assertEqual(取日期('2020.01.05'), Timestamp(2020,1,5))
        self.assertEqual(取日期('20200105'), Timestamp(2020,1,5))
        self.assertEqual(取日期('111/07/01 02:22:34'), Timestamp(2022,7,1))
        self.assertEqual(取日期('110/12/27'), Timestamp(2021,12,27))
        self.assertEqual(取日期('88/02/01'), Timestamp(1999,2,1))
        self.assertEqual(取日期('110.11.10'), Timestamp(2021,11,10))
        self.assertEqual(取日期('111.9.23'), Timestamp(2022,9,23))
        self.assertEqual(取日期('920526'), Timestamp(2003,5,26))
        self.assertEqual(取日期(11204), Timestamp(2023,4,30))
        self.assertEqual(取日期('民國 99 年 09 月 10 日'), Timestamp(2010,9,10))
        self.assertEqual(取日期('109年 1月'), Timestamp(2020,1,31))
        self.assertEqual(取日期('111.4.29+150'), Timestamp(2022,9,26))

        self.assertTrue(pd.isnull(取日期('民國 0 年 00 月 00 日')))

        now = Timestamp.now()
        self.assertEqual(取日期('12.24'), Timestamp(now.year,12,24))
        self.assertEqual(取日期('0831'), Timestamp(now.year, 8, 31))

        self.assertEqual(取日期(datetime(2022,2,1,23,11)), Timestamp(2022,2,1))
        self.assertEqual(取日期('1110527'), Timestamp(2022,5,27))

        self.assertEqual(取日期('1110213.000000'), Timestamp(2022,2,13))
        self.assertEqual(取日期(940101), Timestamp(2005,1,1))
        self.assertEqual(取日期(1110213), Timestamp(2022,2,13))
        self.assertEqual(取日期(1110213.0), Timestamp(2022,2,13))
        self.assertEqual(取日期('昨日'), Timestamp.today().normalize() - Timedelta(days=1))
        self.assertEqual(取日期(date(2022,7,29)), Timestamp(2022,7,29))

    def test_date_predicate(self):
        from zhongwen.date import 取日期
        from zhongwen.date import 全是日期嗎, 有日期嗎
        import pandas as pd

        self.assertFalse(全是日期嗎(pd.NaT))
        self.assertFalse(全是日期嗎([取日期('7.29'), pd.NaT]))
        self.assertTrue(全是日期嗎(取日期('7.29')))
        self.assertTrue(有日期嗎([取日期('7.29'), pd.NaT]))

    @patch('zhongwen.date.今日')
    def test_date_repr(self, 仿今日):
        from zhongwen.date import 民國日期, 取日期, 民國正式日期
        from datetime import datetime
        import pandas as pd

        self.assertEqual(民國日期(取日期('110/12/27')), '1101227')
        self.assertEqual(民國日期(取日期('110/1/27'), '%Y年%M月%d日'), '110年1月27日')
        self.assertEqual(民國日期(pd.NaT), '')

        from zhongwen.date import 經過日數
        self.assertEqual(經過日數('108.5.13', '109.12.3'), 570)

        self.assertEqual(今日(), 取日期(datetime.now()))

        仿今日.return_value = 取日期('113.4.6')
        self.assertEqual(民國日期(取日期('113.4.6'), '%Y年%M月%D日'), '113年4月6日')
        self.assertEqual(民國正式日期(), '113年4月6日')


    @patch('zhongwen.date.今日')
    def test_days(self, 仿今日):
        from zhongwen.date import 取日期, 前天, 近三日
        仿今日.return_value = 取日期('113.1.1')
        self.assertEqual(前天(), 取日期('112.12.30'))
        self.assertIn(取日期('112.12.30'), 近三日())
        self.assertIn(取日期('112.12.31'), 近三日())
        self.assertIn(取日期('113.1.1'), 近三日())

    @patch('zhongwen.date.今日')
    def testmonth(self, 仿今日):
        from zhongwen.date import 取日期, 月底, 上月
        from zhongwen.date import 月起迄, 自起日按日列舉迄今
        from pandas import Timestamp

        self.assertEqual(月底(2022, 10), Timestamp(2022, 10, 31))

        仿今日.return_value = 取日期('113.1.14')
        self.assertEqual(上月(), Timestamp(2023, 12, 31))

        self.assertEqual(月起迄(2022, 4), [Timestamp(2022, 4, 1), Timestamp(2022, 4, 30)])
        dates = list(自起日按日列舉迄今(取日期('113.1.1')))
        self.assertEqual(len(dates), 14)
        self.assertIn(取日期('113.1.1'), dates)
        self.assertIn(取日期('113.1.2'), dates)
        self.assertIn(取日期('113.1.14'), dates)

    def test_quarter(self):
        from zhongwen.date import 季末, 季初, 季別, 與季末相距月數, 迄每季, 取日期
        from pandas import Timestamp
        
        self.assertEqual(季末(2022, 2), Timestamp(2022, 6, 30))

        self.assertEqual(季別(Timestamp(2023, 1, 2)), (2023, 1))
        self.assertEqual(季別(Timestamp(2024, 7, 26)), (2024, 3))
        self.assertEqual(與季末相距月數(Timestamp(2023, 5, 3)), 1)

        self.assertIsNotNone(季初()) 
        qs = 迄每季(取日期('20190101'))
        self.assertEqual(next(qs), 取日期('20190331'))
        self.assertEqual(next(qs), 取日期('20190630'))
        # self.assertEqual(qs[0], 取日期('1120813'))

    def test2(self):
        from zhongwen.date import 民國年底, 取日期, 自起算民國年逐年列舉迄今年, 取日期
        self.assertEqual(民國年底(110), 取日期('1101231'))
        self.assertEqual(list(自起算民國年逐年列舉迄今年(110))[:2]
                        ,[民國年底(110), 民國年底(111)]
                        )

    def test_year_apart(self):
        from zhongwen.date import 相距年數, 取日期
        self.assertEqual(相距年數(取日期('113.9.15'), 取日期('112.9.15')),  1)
        self.assertEqual(相距年數(取日期('113.9.15'), 取日期('112.9.16')),  0)
        self.assertEqual(相距年數(取日期('113.9.15'), 取日期('112.10.1')),  0)

if __name__ == '__main__':
    unittest.main()
